# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np
from PIL import Image
from qtpy import QtWidgets, QtCore
import pyqtgraph as pg
from xypattern import Pattern

from ...widgets.UtilityWidgets import open_file_dialog, open_files_dialog, save_file_dialog
# imports for type hinting in PyCharm -- DO NOT DELETE
from ...widgets.integration import IntegrationWidget
from ...model.DioptasModel import DioptasModel
from ...model.util.HelperModule import get_partial_index, get_partial_value

from .EpicsController import EpicsController


class ImageController(object):
    """
    The ImageController manages the Image actions in the Integration Window. It connects the file actions, as
    well as interaction with the image_view.
    """

    def __init__(self, widget: IntegrationWidget, dioptas_model: DioptasModel):
        """
        :param widget: Reference to IntegrationWidget
        :param dioptas_model: Reference to DioptasModel object
        """
        self.widget = widget
        self.model = dioptas_model

        self.epics_controller = EpicsController(self.widget, self.model)

        self.img_docked = True
        self.view_mode = 'normal'  # modes available: normal, alternative
        self.roi_active = False

        self.vertical_splitter_alternative_state = None
        self.vertical_splitter_normal_state = None
        self.horizontal_splitter_alternative_state = None
        self.horizontal_splitter_normal_state = None

        self.create_signals()
        self.create_mouse_behavior()

    def plot_img(self, auto_scale=None):
        """
        Plots the current image loaded in self.img_data.
        :param auto_scale:
            Determines if intensities should be auto-scaled. If value is None it will use the parameter saved in the
            Object (self._auto_scale)
        """
        if auto_scale is None:
            auto_scale = self.widget.img_autoscale_btn.isChecked()

        if self.widget.integration_image_widget.show_background_subtracted_img_btn.isChecked():
            self.widget.img_widget.plot_image(self.model.img_model.img_data, False)
        else:
            self.widget.img_widget.plot_image(self.model.img_model.raw_img_data, False)

        if auto_scale:
            self.widget.img_widget.auto_level()

    def plot_cake(self, auto_scale=None):
        """
        Plots the cake saved in the calibration data
        :param auto_scale:
            Determines if the intensity should be auto-scaled. If value is None it will use the parameter saved in the
            object (self._auto_scale)
        """
        if auto_scale is None:
            auto_scale = self.widget.img_autoscale_btn.isChecked()

        shift_amount = self.widget.cake_shift_azimuth_sl.value()
        self.widget.cake_widget.plot_image(np.roll(self.model.cake_data, shift_amount, axis=0))
        self.plot_cake_integral()
        self.update_cake_axes_range()
        if auto_scale:
            self.widget.cake_widget.auto_level()

    def plot_cake_integral(self, tth=None):
        if not self.widget.cake_widget.cake_integral_plot.isVisible():
            return

        if tth is None:
            tth = self.model.clicked_tth

        x, y = self.model.calibration_model.cake_integral(
            tth,
            self.widget.integration_control_widget.integration_options_widget.cake_integral_width_sb.value()
        )
        shift_amount = self.widget.cake_shift_azimuth_sl.value()
        self.widget.cake_widget.plot_cake_integral(x, np.roll(y, shift_amount))

    def _update_cake_integral(self):
        self.plot_cake_integral()

    def save_cake_integral(self):
        img_filename, _ = os.path.splitext(os.path.basename(self.model.img_model.filename))
        filename = save_file_dialog(
            self.widget, "Save Cake Integral Data.",
            os.path.join(self.model.working_directories['pattern'],
                         img_filename + '.xy'))

        if filename != '':
            integral_pattern = Pattern(*self.widget.cake_widget.cake_integral_item.getData())
            integral_pattern.save(filename)

    def plot_mask(self):
        """
        Plots the mask data.
        """
        if self.model.use_mask and self.widget.img_mode == 'Image':
            self.widget.img_widget.activate_mask()
            self.widget.img_widget.plot_mask(self.model.mask_model.get_img())
        else:
            self.widget.img_widget.deactivate_mask()

    def update_mask_transparency(self):
        """
        Changes the colormap of the mask according to the transparency option selection in the GUI. Resulting Mask will
        be either transparent or solid.
        """
        self.model.transparent_mask = self.widget.mask_transparent_cb.isChecked()
        if self.model.transparent_mask:
            self.widget.img_widget.set_mask_color([255, 0, 0, 100])
        else:
            self.widget.img_widget.set_mask_color([255, 0, 0, 255])

    def create_signals(self):
        self.model.configuration_selected.connect(self.update_gui_from_configuration)
        self.model.img_changed.connect(self.update_img_control_widget)

        self.model.img_changed.connect(self.plot_img)
        self.model.img_changed.connect(self.plot_mask)

        """
        Creates all the connections of the GUI elements.
        """
        self.widget.img_step_file_widget.next_btn.clicked.connect(self.load_next_img)
        self.widget.img_step_file_widget.previous_btn.clicked.connect(self.load_previous_img)
        self.widget.load_img_btn.clicked.connect(self.load_file)
        self.widget.img_filename_txt.editingFinished.connect(self.filename_txt_changed)
        self.widget.img_directory_txt.editingFinished.connect(self.directory_txt_changed)
        self.widget.img_directory_btn.clicked.connect(self.img_directory_btn_click)

        self.widget.img_step_series_widget.next_btn.clicked.connect(self.load_next_series_img)
        self.widget.img_step_series_widget.previous_btn.clicked.connect(self.load_prev_series_img)
        self.widget.img_step_series_widget.pos_txt.editingFinished.connect(self.load_series_img)

        self.widget.file_info_btn.clicked.connect(self.show_file_info)

        self.widget.img_step_file_widget.browse_by_name_rb.clicked.connect(self.set_iteration_mode_number)
        self.widget.img_step_file_widget.browse_by_time_rb.clicked.connect(self.set_iteration_mode_time)

        self.widget.image_control_widget.sources_cb.currentTextChanged.connect(self.select_source)

        ###
        # Image widget image specific controls
        self.widget.img_roi_btn.clicked.connect(self.click_roi_btn)
        self.widget.img_mask_btn.clicked.connect(self.change_mask_mode)
        self.widget.mask_transparent_cb.clicked.connect(self.update_mask_transparency)

        ###
        # Image Widget cake specific controls
        self.widget.img_phases_btn.clicked.connect(self.toggle_show_phases)
        self.widget.cake_shift_azimuth_sl.valueChanged.connect(self.cake_shift_changed)
        self.widget.integration_image_widget.cake_view.img_view_box.sigRangeChanged.connect(self.update_cake_axes_range)
        self.widget.pattern_q_btn.clicked.connect(self.set_cake_axis_to_q)
        self.widget.pattern_tth_btn.clicked.connect(self.set_cake_axis_to_2th)
        self.widget.integration_control_widget.integration_options_widget.cake_integral_width_sb.valueChanged. \
            connect(self._update_cake_integral)
        self.widget.integration_control_widget.integration_options_widget.cake_integral_width_sb.editingFinished. \
            connect(self._update_cake_integral)
        self.widget.integration_control_widget.integration_options_widget.cake_save_integral_btn.clicked. \
            connect(self.save_cake_integral)

        ###
        # General Image Widget controls
        self.widget.img_dock_btn.clicked.connect(self.img_dock_btn_clicked)
        self.widget.img_autoscale_btn.clicked.connect(self.img_autoscale_btn_clicked)
        self.widget.img_mode_btn.clicked.connect(self.change_view_mode)

        self.widget.integration_image_widget.show_background_subtracted_img_btn.clicked.connect(
            self.show_background_subtracted_img_btn_clicked)

        self.widget.qa_save_img_btn.clicked.connect(self.save_img)
        self.widget.load_calibration_btn.clicked.connect(self.load_calibration)
        self.widget.set_wavelnegth_btn.clicked.connect(self.set_wavelength)

        # signals
        self.widget.change_view_btn.clicked.connect(self.change_view_btn_clicked)
        self.widget.autoprocess_cb.toggled.connect(self.auto_process_cb_click)

    def activate(self):
        if self.widget.img_mode == 'Image':
            if not self.model.img_changed.has_listener(self.plot_img):
                self.model.img_changed.connect(self.plot_img)
            if not self.model.img_changed.has_listener(self.plot_mask):
                self.model.img_changed.connect(self.plot_mask)
        elif self.widget.img_mode == 'Cake':
            if not self.model.cake_changed.has_listener(self.plot_cake):
                self.model.cake_changed.connect(self.plot_cake)
        self.widget.calibration_lbl.setText(
            self.model.calibration_model.calibration_name
        )
        self.widget.wavelength_lbl.setText(
            "{:.4f}".format(self.model.calibration_model.wavelength * 1e10) + " A"
        )

    def update_image(self):
        if self.widget.img_mode == 'Image':
            self.plot_mask()
            self.plot_img()
        elif self.widget.img_mode == 'Cake':
            self.plot_cake()

    def deactivate(self):
        if self.model.img_changed.has_listener(self.plot_img):
            self.model.img_changed.disconnect(self.plot_img)
        if self.plot_mask in self.model.img_changed.listeners:
            self.model.img_changed.disconnect(self.plot_mask)
        if self.plot_cake in self.model.cake_changed.listeners:
            self.model.cake_changed.disconnect(self.plot_cake)

    def create_mouse_behavior(self):
        """
        Creates the signal connections of mouse interactions
        """
        self.widget.img_widget.mouse_left_clicked.connect(self.img_mouse_click)
        self.widget.img_widget.mouse_moved.connect(self.show_img_mouse_position)
        self.widget.cake_widget.mouse_left_clicked.connect(self.img_mouse_click)
        self.widget.cake_widget.mouse_moved.connect(self.show_img_mouse_position)

        self.model.clicked_tth_changed.connect(self.set_cake_line_position)
        self.model.clicked_tth_changed.connect(self.set_image_line_position)

    def load_file(self, *args, **kwargs):
        filename = kwargs.get('filename', None)
        if filename is None:
            filenames = open_files_dialog(self.widget, "Load image data file(s)",
                                          self.model.working_directories['image'])
        else:
            filenames = [filename]

        if filenames is not None and len(filenames) != 0:
            self.model.working_directories['image'] = os.path.dirname(str(filenames[0]))
            if len(filenames) == 1:
                self.model.img_model.load(str(filenames[0]))
            else:
                if self.widget.img_batch_mode_add_rb.isChecked():
                    self.model.img_model.blockSignals(True)
                    self.model.img_model.load(str(filenames[0]))
                    for ind in range(1, len(filenames)):
                        self.model.img_model.add(filenames[ind])
                    self.model.img_model.blockSignals(False)
                    self.model.img_model.img_changed.emit()
                if self.widget.img_batch_mode_average_rb.isChecked():
                    self.model.img_model.blockSignals(True)
                    self.model.img_model.load(str(filenames[0]))
                    for ind in range(1, len(filenames)):
                        self.model.img_model.add(filenames[ind])
                    self.model.img_model._img_data = self.model.img_model._img_data / len(filenames)
                    self.model.img_model._calculate_img_data()
                    self.model.img_model.blockSignals(False)
                    self.model.img_model.img_changed.emit()
                elif self.widget.img_batch_mode_integrate_rb.isChecked():
                    self._load_multiple_files(filenames)
                elif self.widget.img_batch_mode_image_save_rb.isChecked():
                    self._save_multiple_image_files(filenames)

    def _load_multiple_files(self, filenames):
        if not self.model.calibration_model.is_calibrated:
            self.widget.show_error_msg("Can not integrate multiple images without calibration.")
            return

        working_directory = self._get_pattern_working_directory()
        if working_directory == '':
            return  # abort file processing if no directory was selected

        progress_dialog = self.widget.get_progress_dialog("Integrating multiple files.", "Abort Integration",
                                                          len(filenames))
        self._set_up_batch_processing()

        for ind in range(len(filenames)):
            filename = str(filenames[ind])
            base_filename = os.path.basename(filename)

            progress_dialog.setValue(ind)
            progress_dialog.setLabelText("Integrating: " + base_filename)

            self.model.img_model.blockSignals(True)
            self.model.img_model.load(filename)
            self.model.img_model.blockSignals(False)

            x, y = self.integrate_pattern()
            self._save_pattern(base_filename, working_directory, x, y)

            QtWidgets.QApplication.processEvents()
            if progress_dialog.wasCanceled():
                break

        progress_dialog.close()
        self._tear_down_batch_processing()

    def _get_pattern_working_directory(self):
        if self.widget.pattern_autocreate_cb.isChecked():
            working_directory = self.model.working_directories['pattern']
        else:
            # if there is no working directory selected A file dialog opens up to choose a directory...
            working_directory = str(QtWidgets.QFileDialog.getExistingDirectory(
                self.widget, "Please choose the output directory for the integrated Patterns.",
                self.model.working_directories['pattern']))
        return working_directory

    def _set_up_batch_processing(self):
        self.model.blockSignals(True)

    def _tear_down_batch_processing(self):
        self.model.blockSignals(False)
        self.model.img_changed.emit()
        self.model.pattern_changed.emit()

    def _save_multiple_image_files(self, filenames):
        working_directory = str(QtWidgets.QFileDialog.getExistingDirectory(
            self.widget, "Please choose the output directory for the Images.",
            self.model.working_directories['image']))

        if working_directory == '':
            return

        self._set_up_batch_processing()
        progress_dialog = self.widget.get_progress_dialog("Saving multiple image files.", "Abort",
                                                          len(filenames))
        QtWidgets.QApplication.processEvents()

        self.model.current_configuration.auto_integrate_pattern = False

        for ind, filename in enumerate(filenames):
            base_filename = os.path.basename(filename)

            progress_dialog.setValue(ind)
            progress_dialog.setLabelText("Saving: " + base_filename)

            self.model.img_model.load(str(filename))
            self.save_img(os.path.join(working_directory, 'batch_' + base_filename))

            QtWidgets.QApplication.processEvents()
            if progress_dialog.wasCanceled():
                break

        self.model.current_configuration.auto_integrate_pattern = True

        progress_dialog.close()
        self._tear_down_batch_processing()

    def _save_pattern(self, base_filename, working_directory, x, y):
        file_endings = self._get_pattern_file_endings()
        for file_ending in file_endings:
            filename = os.path.join(working_directory, os.path.splitext(base_filename)[0] + file_ending)
            self.model.pattern_model.set_pattern(x, y, filename, unit=self.get_integration_unit())
            if file_ending == '.xy':
                self.model.pattern_model.save_pattern(filename, header=self._create_pattern_header())
            else:
                self.model.pattern_model.save_pattern(filename)

            # save the background subtracted filename
            if self.model.pattern.auto_bkg is not None:
                directory = os.path.join(working_directory, 'bkg_subtracted')
                if not os.path.exists(directory):
                    os.mkdir(directory)
                filename = os.path.join(directory, self.model.pattern.name + file_ending)
                if file_ending == '.xy':
                    self.model.pattern_model.save_pattern(filename, header=self._create_pattern_header(),
                                                          subtract_background=True)
                else:
                    self.model.pattern_model.save_pattern(filename, subtract_background=True)

    def _create_pattern_header(self):
        header = self.model.calibration_model.create_file_header()
        header = header.replace('\r\n', '\n')
        header += '\n#\n# ' + self.model.pattern_model.unit + '\t I'
        return header

    def _get_pattern_file_endings(self):
        res = []
        if self.widget.pattern_header_xy_cb.isChecked():
            res.append('.xy')
        if self.widget.pattern_header_chi_cb.isChecked():
            res.append('.chi')
        if self.widget.pattern_header_dat_cb.isChecked():
            res.append('.dat')
        return res

    def show_file_info(self):
        self.widget.file_info_widget.raise_widget()

    def get_integration_unit(self):
        if self.widget.pattern_tth_btn.isChecked():
            return '2th_deg'
        elif self.widget.pattern_q_btn.isChecked():
            return 'q_A^-1'
        elif self.widget.pattern_d_btn.isChecked():
            return 'd_A'

    def integrate_pattern(self):
        if self.widget.img_mask_btn.isChecked():
            mask = self.model.mask_model.get_mask()
        else:
            mask = None

        if self.widget.img_roi_btn.isChecked():
            roi_mask = self.widget.img_widget.roi.getRoiMask(self.model.img_data.shape)
        else:
            roi_mask = None

        if roi_mask is None and mask is None:
            mask = None
        elif roi_mask is None and mask is not None:
            mask = mask
        elif roi_mask is not None and mask is None:
            mask = roi_mask
        elif roi_mask is not None and mask is not None:
            mask = np.logical_or(mask, roi_mask)

        if self.widget.pattern_tth_btn.isChecked():
            integration_unit = '2th_deg'
        elif self.widget.pattern_q_btn.isChecked():
            integration_unit = 'q_A^-1'
        elif self.widget.pattern_d_btn.isChecked():
            integration_unit = 'd_A'
        else:
            # in case something weird happened
            print('No correct integration unit selected')
            return

        if not self.widget.automatic_binning_cb.isChecked():
            num_points = int(str(self.widget.bin_count_txt.text()))
        else:
            num_points = None
        return self.model.calibration_model.integrate_1d(mask=mask, unit=integration_unit, num_points=num_points)

    def change_mask_mode(self):
        self.model.use_mask = self.widget.integration_image_widget.mask_btn.isChecked()
        self.widget.mask_transparent_cb.setVisible(self.model.use_mask)
        self.plot_mask()
        self.model.img_model.img_changed.emit()

    def update_mask_mode(self):
        self.widget.integration_image_widget.mask_btn.setChecked(bool(self.model.use_mask))
        self.widget.mask_transparent_cb.setVisible(bool(self.model.use_mask))
        self.widget.mask_transparent_cb.setChecked(bool(self.model.transparent_mask))

    def update_img_mode(self):
        self.widget.img_mode_btn.click()

    def load_series_img(self):
        pos = int(str(self.widget.img_step_series_widget.pos_txt.text()))
        self.model.img_model.load_series_img(pos)

    def load_prev_series_img(self):
        step = int(str(self.widget.img_step_series_widget.step_txt.text()))
        pos = int(str(self.widget.img_step_series_widget.pos_txt.text()))
        self.model.img_model.load_series_img(pos - step)

    def load_next_series_img(self):
        step = int(str(self.widget.img_step_series_widget.step_txt.text()))
        pos = int(str(self.widget.img_step_series_widget.pos_txt.text()))
        self.model.img_model.load_series_img(pos + step)

    def load_next_img(self):
        step = int(str(self.widget.img_step_file_widget.step_txt.text()))
        self.model.img_model.load_next_file(step=step)

    def load_previous_img(self):
        step = int(str(self.widget.img_step_file_widget.step_txt.text()))
        self.model.img_model.load_previous_file(step=step)

    def filename_txt_changed(self):
        current_filename = os.path.basename(self.model.img_model.filename)
        current_directory = str(self.widget.img_directory_txt.text())
        new_filename = str(self.widget.img_filename_txt.text())
        if os.path.exists(os.path.join(current_directory, new_filename)):
            try:
                self.load_file(filename=os.path.join(current_directory, new_filename).replace('\\', '/'))
            except TypeError:
                self.widget.img_filename_txt.setText(current_filename)
        else:
            self.widget.img_filename_txt.setText(current_filename)

    def directory_txt_changed(self):
        new_directory = str(self.widget.img_directory_txt.text())
        if os.path.exists(new_directory) and new_directory != self.model.working_directories['image']:
            if self.model.img_model.autoprocess:
                self._files_now = dict([(f, None) for f in os.listdir(self.model.working_directories['image'])])
            self.model.working_directories['image'] = os.path.abspath(new_directory)
            old_filename = str(self.widget.img_filename_txt.text())
            self.widget.img_filename_txt.setText(old_filename + '*')
        else:
            self.widget.img_directory_txt.setText(self.model.working_directories['image'])

    def img_directory_btn_click(self):
        directory = str(QtWidgets.QFileDialog.getExistingDirectory(
            self.widget,
            "Please choose the image working directory.",
            self.model.working_directories['image']))
        if directory != '':
            if self.model.img_model.autoprocess:
                self._files_now = dict([(f, None) for f in os.listdir(self.model.working_directories['image'])])
            self.model.working_directories['image'] = directory
            self.widget.img_directory_txt.setText(directory)

    def update_img_control_widget(self):
        self.widget.img_step_series_widget.setVisible(int(self.model.img_model.series_max > 1))
        self.widget.img_step_series_widget.pos_validator.setTop(self.model.img_model.series_max)
        self.widget.img_step_series_widget.pos_txt.setText(str(self.model.img_model.series_pos))

        self.widget.file_info_btn.setVisible(self.model.img_model.file_info != "")
        self.widget.move_btn.setVisible(len(self.model.img_model.motors_info) > 0)

        self.widget.img_filename_txt.setText(os.path.basename(self.model.img_model.filename))
        self.widget.img_directory_txt.setText(os.path.dirname(self.model.img_model.filename))
        self.widget.file_info_widget.text_lbl.setText(self.model.img_model.file_info)

        self.widget.image_control_widget.sources_widget.setVisible(not (self.model.img_model.sources is None))
        if self.model.img_model.sources is not None:
            sources_cb = self.widget.image_control_widget.sources_cb
            sources_cb.blockSignals(True)
            # remove all previous items:
            for _ in range(sources_cb.count()):
                sources_cb.removeItem(0)

            sources_cb.addItems(self.model.img_model.sources)
            sources_cb.setCurrentText(self.model.img_model.selected_source)
            sources_cb.blockSignals(False)

        self.widget.cbn_plot_btn.setText('Plot')
        self.widget.oiadac_plot_btn.setText('Plot')

        # update the window due to some errors on mac when using macports
        self._get_master_parent().update()

    def _get_master_parent(self):
        master_widget_parent = self.widget
        while master_widget_parent.parent():
            master_widget_parent = master_widget_parent.parent()
        return master_widget_parent

    def click_roi_btn(self):
        if self.model.current_configuration.roi is None:
            self.model.current_configuration.roi = self.widget.img_widget.roi.getRoiLimits()
        else:
            self.model.current_configuration.roi = None
        self.update_roi_in_gui()

    def update_roi_in_gui(self):
        roi = self.model.mask_model.roi
        if roi is None:
            self.widget.img_widget.deactivate_roi()
            self.widget.img_roi_btn.setChecked(False)
            if self.roi_active:
                self.widget.img_widget.roi.sigRegionChangeFinished.disconnect(self.update_roi_in_model)
                self.roi_active = False
            return

        if not self.model.current_configuration.auto_integrate_cake:
            self.widget.img_roi_btn.setChecked(True)
            self.widget.img_widget.activate_roi()
            self.widget.img_widget.update_roi_shade_limits(self.model.img_data.shape)

            pos = QtCore.QPoint(int(roi[2]), int(roi[0]))
            size = QtCore.QPoint(int(roi[3] - roi[2]), int(roi[1] - roi[0]))
            self.widget.img_widget.roi.setRoiLimits(pos, size)

            if not self.roi_active:
                self.widget.img_widget.roi.sigRegionChangeFinished.connect(self.update_roi_in_model)
                self.roi_active = True

    def update_roi_in_model(self):
        self.model.current_configuration.roi = self.widget.img_widget.roi.getRoiLimits()

    def change_view_mode(self):
        if str(self.widget.img_mode_btn.text()) == 'Cake':
            self.activate_cake_mode()
        elif str(self.widget.img_mode_btn.text()) == 'Image':
            self.activate_image_mode()

    def toggle_show_phases(self):
        if str(self.widget.img_phases_btn.text()) == 'Show Phases':
            self.widget.integration_image_widget.cake_view.show_all_visible_cake_phases(
                self.widget.phase_widget.phase_show_cbs)
            self.widget.img_phases_btn.setText('Hide Phases')
            self.model.enabled_phases_in_cake.emit()
        elif str(self.widget.img_phases_btn.text()) == 'Hide Phases':
            self.widget.integration_image_widget.cake_view.hide_all_cake_phases()
            self.widget.img_phases_btn.setText('Show Phases')

    def cake_shift_changed(self):
        self.plot_cake()
        self._update_cake_mouse_click_pos()
        self.update_cake_azimuth_axis()
        self.plot_cake_integral(None)

    def activate_cake_mode(self):
        if self.model.calibration_model.cake_geometry is None:
            self.widget.show_error_msg("Can not switch to the cake mode without calibration.")
            return

        if not self.model.current_configuration.auto_integrate_cake:
            self.model.current_configuration.auto_integrate_cake = True

        self.model.current_configuration.integrate_image_2d()

        self.set_cake_line_position(self.model.clicked_tth)
        self._update_cake_mouse_click_pos()

        self.widget.img_mode_btn.setText('Image')
        self.widget.img_mode = str("Cake")

        self.model.img_changed.disconnect(self.plot_img)
        self.model.img_changed.disconnect(self.plot_mask)

        self.model.cake_changed.connect(self.plot_cake)
        self.plot_cake()

        self.widget.cake_shift_azimuth_sl.setVisible(True)
        self.widget.cake_shift_azimuth_sl.setMinimum(int(-len(self.model.cake_azi) / 2))
        self.widget.cake_shift_azimuth_sl.setMaximum(int(len(self.model.cake_azi) / 2))
        self.widget.cake_shift_azimuth_sl.setSingleStep(1)
        self.widget.img_phases_btn.setVisible(True)

        self.widget.integration_image_widget.img_pg_layout.hide()
        self.widget.integration_image_widget.cake_pg_layout.show()

    def activate_image_mode(self):
        if self.model.current_configuration.auto_integrate_cake:
            self.model.current_configuration.auto_integrate_cake = False

        self.widget.cake_shift_azimuth_sl.setVisible(False)
        self.widget.img_phases_btn.setVisible(False)

        self._update_image_line_pos()
        self._update_image_mouse_click_pos()

        self.widget.img_mode_btn.setText('Cake')
        self.widget.img_mode = str("Image")
        self.model.img_changed.connect(self.plot_img)
        self.model.img_changed.connect(self.plot_mask)
        self.model.cake_changed.disconnect(self.plot_cake)
        self.plot_img()
        self.plot_mask()

        self.widget.integration_image_widget.img_pg_layout.show()
        self.widget.integration_image_widget.cake_pg_layout.hide()

    def img_autoscale_btn_clicked(self):
        if self.widget.img_autoscale_btn.isChecked():
            self.widget.img_widget.auto_level()

    def img_dock_btn_clicked(self):
        self.img_docked = not self.img_docked
        self.widget.dock_img(self.img_docked)

    def show_background_subtracted_img_btn_clicked(self):
        if self.widget.img_mode_btn.text() == 'Cake':
            self.plot_img()
        else:
            self.widget.integration_image_widget.show_background_subtracted_img_btn.setChecked(False)

    def _update_cake_mouse_click_pos(self):
        if not self.model.calibration_model.is_calibrated:
            return

        tth = self.model.clicked_tth
        azi = self.model.clicked_azi

        cake_tth = self.model.cake_tth

        x_pos = get_partial_index(cake_tth, tth)
        if x_pos is None:
            return

        x_pos = x_pos + 0.5
        shift_amount = self.widget.cake_shift_azimuth_sl.value()
        y_pos = (get_partial_index(self.model.cake_azi, azi) + 0.5 + shift_amount) % len(self.model.cake_azi)
        self.widget.cake_widget.set_mouse_click_position(x_pos, y_pos)

    def _update_image_line_pos(self):
        if not self.model.calibration_model.is_calibrated:
            return
        self.set_image_line_position(self.model.clicked_tth)

    def _update_image_mouse_click_pos(self):
        if not self.model.calibration_model.is_calibrated:
            return

        tth = np.deg2rad(self.model.clicked_tth)
        azi = np.deg2rad(self.model.clicked_azi)

        new_pos = self.model.calibration_model.get_pixel_ind(tth, azi)
        if len(new_pos) == 0:
            self.widget.img_widget.mouse_click_item.hide()
        else:
            x_ind, y_ind = new_pos
            self.widget.img_widget.set_mouse_click_position(y_ind + 0.5, x_ind + 0.5)
            self.widget.img_widget.mouse_click_item.show()

    def update_cake_axes_range(self):
        if self.model.current_configuration.auto_integrate_cake:
            self.update_cake_azimuth_axis()
            self.update_cake_x_axis()

    def update_cake_azimuth_axis(self):
        img_view_rect = self.widget.integration_image_widget.cake_view.img_view_rect()
        img_bounding_rect = self.widget.integration_image_widget.cake_view.img_bounding_rect()
        shift_amount = self.widget.cake_shift_azimuth_sl.value()
        cake_azi = self.model.cake_azi - shift_amount * np.mean(np.diff(self.model.cake_azi))
        if img_bounding_rect.height() == 0:
            return

        height = img_view_rect.height()
        bottom = img_view_rect.top()
        v_scale = (cake_azi[-1] - cake_azi[0]) / img_bounding_rect.height()
        v_shift = np.min(cake_azi[0])
        min_azi = v_scale * bottom + v_shift
        max_azi = v_scale * (bottom + height) + v_shift

        self.widget.integration_image_widget.cake_view.left_axis_cake.setRange(min_azi, max_azi)

    def update_cake_x_axis(self):
        if self.model.cake_tth is None:
            return

        img_view_rect = self.widget.integration_image_widget.cake_view.img_view_rect()
        img_bounding_rect = self.widget.integration_image_widget.cake_view.img_bounding_rect()

        cake_tth = self.model.cake_tth
        if img_bounding_rect.width() == 0:
            return

        width = img_view_rect.width()
        left = img_view_rect.left()
        h_scale = (np.max(cake_tth) - np.min(cake_tth)) / img_bounding_rect.width()
        h_shift = np.min(cake_tth)
        min_tth = h_scale * left + h_shift
        max_tth = h_scale * (left + width) + h_shift

        if self.model.current_configuration.integration_unit == '2th_deg':
            self.widget.integration_image_widget.cake_view.bottom_axis_cake.setRange(min_tth, max_tth)
        elif self.model.current_configuration.integration_unit == 'q_A^-1':
            self.widget.integration_image_widget.cake_view.bottom_axis_cake.setRange(
                self.convert_x_value(min_tth, '2th_deg', 'q_A^-1'),
                self.convert_x_value(max_tth, '2th_deg', 'q_A^-1'))

    def set_cake_axis_to_2th(self):
        self.widget.integration_image_widget.cake_view.bottom_axis_cake.setLabel(u'2θ', u'°')
        self.update_cake_x_axis()

    def set_cake_axis_to_q(self):
        self.widget.integration_image_widget.cake_view.bottom_axis_cake.setLabel('Q', 'A<sup>-1</sup>')
        self.update_cake_x_axis()

    def show_img_mouse_position(self, x, y):
        if self.widget.img_mode == 'Cake':
            img_data = self.widget.cake_widget.img_data
        else:
            img_data = self.widget.img_widget.img_data
        img_shape = img_data.shape

        if 0 < x < img_shape[1] - 1 and 0 < y < img_shape[0] - 1:
            self.update_mouse_position_labels(x, y, img_data[int(np.floor(y)), int(np.floor(x))])
        else:
            self.update_mouse_position_labels(x, y, None)

        if 0.5 < x < img_shape[1] - 0.5 and 0.5 < y < img_shape[0] - 0.5:
            if self.model.calibration_model.is_calibrated:
                x_temp = x
                x = np.array([y])
                y = np.array([x_temp])
                if self.widget.img_mode == 'Cake':
                    tth = get_partial_value(self.model.cake_tth, y - 0.5)
                    shift_amount = self.widget.cake_shift_azimuth_sl.value()
                    cake_azi = self.model.cake_azi - shift_amount * np.mean(np.diff(self.model.cake_azi))
                    azi = get_partial_value(cake_azi, x - 0.5)
                    q_value = self.convert_x_value(tth, '2th_deg', 'q_A^-1')
                else:
                    tth = self.model.calibration_model.get_two_theta_img(x, y)
                    tth = tth / np.pi * 180.0
                    q_value = self.convert_x_value(tth, '2th_deg', 'q_A^-1')
                    azi = self.model.calibration_model.get_azi_img(x, y) / np.pi * 180

                d = self.convert_x_value(tth, '2th_deg', 'd_A')
                tth_str = u"2θ:%9.3f  " % tth
                self.widget.mouse_tth_lbl.setText(tth_str)
                self.widget.mouse_d_lbl.setText('d:%9.3f  ' % d)
                self.widget.mouse_q_lbl.setText('Q:%9.3f  ' % q_value)
                self.widget.mouse_azi_lbl.setText('X:%9.3f  ' % azi)
                self.widget.img_widget_mouse_tth_lbl.setText(tth_str)
                self.widget.img_widget_mouse_d_lbl.setText('d:%9.3f  ' % d)
                self.widget.img_widget_mouse_q_lbl.setText('Q:%9.3f  ' % q_value)
                self.widget.img_widget_mouse_azi_lbl.setText('X:%9.3f  ' % azi)
            else:
                self.widget.mouse_tth_lbl.setText(u'2θ: -')
                self.widget.mouse_d_lbl.setText('d: -')
                self.widget.mouse_q_lbl.setText('Q: -')
                self.widget.mouse_azi_lbl.setText('X: -')
                self.widget.img_widget_mouse_tth_lbl.setText(u'2θ: -')
                self.widget.img_widget_mouse_d_lbl.setText('d: -')
                self.widget.img_widget_mouse_q_lbl.setText('Q: -')
                self.widget.img_widget_mouse_azi_lbl.setText('X: -')

    def img_mouse_click(self, x, y):
        if self.widget.img_mode == 'Cake':
            img_data = self.widget.cake_widget.img_data
        else:
            img_data = self.widget.img_widget.img_data

        if 0 < x < img_data.shape[1] - 1 and 0 < y < img_data.shape[0] - 1:
            intensity = img_data[int(np.floor(y)), int(np.floor(x))]
        else:
            intensity = None
        self.update_mouse_click_position_labels(x, y, intensity)

        if self.model.calibration_model.is_calibrated:
            x, y = y, x  # the indices are reversed for the img_array
            if self.widget.img_mode == 'Cake':  # cake mode
                # get clicked tth and azimuth
                cake_shape = self.model.cake_data.shape
                if x < 0 or y < 0 or x > cake_shape[0] - 1 or y > cake_shape[1] - 1:
                    return
                tth = get_partial_value(self.model.cake_tth, y - 0.5)
                shift_amount = self.widget.cake_shift_azimuth_sl.value()
                azi = get_partial_value(np.roll(self.model.cake_azi, shift_amount), x - 0.5)
                self.widget.cake_widget.activate_vertical_line()

            elif self.widget.img_mode == 'Image':  # image mode
                img_shape = self.model.img_data.shape
                if x < 0 or y < 0 or x > img_shape[0] - 1 or y > img_shape[1] - 1:
                    return
                x = np.array([x])
                y = np.array([y])
                tth = np.rad2deg(self.model.calibration_model.get_two_theta_img(x, y))
                azi = np.rad2deg(self.model.calibration_model.get_azi_img(x, y))
            else:  # in the case of whatever
                tth = 0
                azi = 0

            self.clicked_tth = tth  # in degree
            self.clicked_azi = azi  # in degree

            if self.widget.img_mode == 'Cake':
                self.plot_cake_integral()

            self.model.clicked_tth_changed.emit(tth)
            self.model.clicked_azi_changed.emit(azi)

        self.widget.click_tth_lbl.setText(self.widget.mouse_tth_lbl.text())
        self.widget.click_d_lbl.setText(self.widget.mouse_d_lbl.text())
        self.widget.click_q_lbl.setText(self.widget.mouse_q_lbl.text())
        self.widget.click_azi_lbl.setText(self.widget.mouse_azi_lbl.text())
        self.widget.img_widget_click_tth_lbl.setText(self.widget.mouse_tth_lbl.text())
        self.widget.img_widget_click_d_lbl.setText(self.widget.mouse_d_lbl.text())
        self.widget.img_widget_click_q_lbl.setText(self.widget.mouse_q_lbl.text())
        self.widget.img_widget_click_azi_lbl.setText(self.widget.mouse_azi_lbl.text())

    def update_mouse_position_labels(self, x, y, intensity):
        x_pos_string = 'X:  %4d' % x
        y_pos_string = 'Y:  %4d' % y

        try:
            int_string = 'I:   %5d' % intensity
        except (ValueError, TypeError, OverflowError):
            int_string = 'I:'

        self.widget.mouse_x_lbl.setText(x_pos_string)
        self.widget.mouse_y_lbl.setText(y_pos_string)
        self.widget.mouse_int_lbl.setText(int_string)

    def update_mouse_click_position_labels(self, x, y, intensity):
        self.update_mouse_position_labels(x, y, intensity)

        self.widget.click_x_lbl.setText(self.widget.mouse_x_lbl.text())
        self.widget.click_y_lbl.setText(self.widget.mouse_y_lbl.text())
        self.widget.click_int_lbl.setText(self.widget.mouse_int_lbl.text())

    def set_cake_line_position(self, tth):
        if self.model.cake_tth is None:
            return

        pos = get_partial_index(self.model.cake_tth, tth)
        if pos is None:
            self.widget.cake_widget.plot_cake_integral(np.array([]), np.array([]))
            self.widget.cake_widget.deactivate_vertical_line()
            return

        self.widget.cake_widget.vertical_line.setValue(pos + 0.5)
        self.widget.cake_widget.activate_vertical_line()
        self.plot_cake_integral(tth)

    def set_image_line_position(self, tth):
        if not self.model.calibration_model.is_calibrated:
            self.widget.img_widget.deactivate_circle_scatter()
            return
        self.widget.img_widget.activate_circle_scatter()
        self.widget.img_widget.set_circle_line(self.model.calibration_model.get_two_theta_array(), np.deg2rad(tth))

    def set_iteration_mode_number(self):
        self.model.img_model.set_file_iteration_mode('number')

    def set_iteration_mode_time(self):
        self.model.img_model.set_file_iteration_mode('time')

    def select_source(self, source):
        self.model.img_model.select_source(source)

    def convert_x_value(self, value, previous_unit, new_unit):
        wavelength = self.model.calibration_model.wavelength
        if previous_unit == '2th_deg':
            tth = value
        elif previous_unit == 'q_A^-1':
            tth = np.arcsin(
                value * 1e10 * wavelength / (4 * np.pi)) * 360 / np.pi
        elif previous_unit == 'd_A':
            tth = 2 * np.arcsin(wavelength / (2 * value * 1e-10)) * 180 / np.pi
        else:
            tth = 0

        if new_unit == '2th_deg':
            res = tth
        elif new_unit == 'q_A^-1':
            res = 4 * np.pi * \
                  np.sin(tth / 360 * np.pi) / \
                  wavelength / 1e10
        elif new_unit == 'd_A':
            res = wavelength / (2 * np.sin(tth / 360 * np.pi)) * 1e10
        else:
            res = 0
        return res

    def load_calibration(self):
        filename = open_file_dialog(
            self.widget, "Load calibration...",
            self.model.working_directories['calibration'],
            '*.poni')
        if filename != '':
            self.model.working_directories['calibration'] = os.path.dirname(filename)
            self.model.calibration_model.load(filename)
            self.widget.calibration_lbl.setText(
                self.model.calibration_model.calibration_name)
            self.widget.wavelength_lbl.setText('{:.4f}'.format(self.model.calibration_model.wavelength * 1e10) + ' A')
            self.model.img_model.img_changed.emit()

    def set_wavelength(self):
        wavelength, ok = QtWidgets.QInputDialog.getText(self.widget, 'Set Wavelength', 'Wavelength in Angstroms:')
        if ok:
            self.model.calibration_model.pattern_geometry.wavelength = float(wavelength) * 1e-10
            self.widget.wavelength_lbl.setText('{:.4f}'.format(self.model.calibration_model.wavelength * 1e10) + ' A')
            self.model.img_model.img_changed.emit()

    def auto_process_cb_click(self):
        self.model.img_model.autoprocess = self.widget.autoprocess_cb.isChecked()

    def save_img(self, filename=None):
        if not filename:
            img_filename = os.path.splitext(os.path.basename(self.model.img_model.filename))[0]
            filename = save_file_dialog(self.widget, "Save Image.",
                                        os.path.join(self.model.working_directories['image'],
                                                     img_filename + '.png'),
                                        ('Image (*.png);;Data (*.tiff);;Text (*.txt)'))

        if filename != '':
            if filename.endswith('.png'):
                if self.widget.img_mode == 'Cake':
                    self.widget.cake_widget.deactivate_vertical_line()
                    self.widget.cake_widget.deactivate_mouse_click_item()
                    QtWidgets.QApplication.processEvents()
                    self.widget.cake_widget.save_img(filename)
                    self.widget.cake_widget.activate_vertical_line()
                    self.widget.cake_widget.activate_mouse_click_item()
                elif self.widget.img_mode == 'Image':
                    self.widget.img_widget.deactivate_circle_scatter()
                    self.widget.img_widget.deactivate_roi()
                    QtWidgets.QApplication.processEvents()
                    self.widget.img_widget.save_img(filename)
                    self.widget.img_widget.activate_circle_scatter()
                    if self.roi_active:
                        self.widget.img_widget.activate_roi()

            elif filename.endswith('.tiff') or filename.endswith('.tif'):
                if self.widget.img_mode == 'Image':
                    im_array = np.int32(self.model.img_data)
                elif self.widget.img_mode == 'Cake':
                    im_array = np.int32(self.model.cake_data)
                im_array = np.flipud(im_array)
                im = Image.fromarray(im_array)
                im.save(filename)
            elif filename.endswith('.txt') or filename.endswith('.csv'):
                if self.widget.img_mode == 'Image':
                    return
                elif self.widget.img_mode == 'Cake':  # saving cake data as a text file for export.
                    with open(filename, 'w') as out_file:  # this is done in an odd and slow way because the headers
                        # should be floats and the data itself int.
                        cake_tth = np.insert(self.model.cake_tth, 0, 0)
                        np.savetxt(out_file, cake_tth[None], fmt='%6.3f')
                        for azi, row in zip(self.model.cake_azi, self.model.cake_data):
                            row_str = " ".join(["{:6.0f}".format(el) for el in row])
                            out_file.write("{:6.2f}".format(azi) + row_str + '\n')

    def update_gui_from_configuration(self):
        self.widget.img_mask_btn.setChecked(int(self.model.use_mask))
        self.widget.mask_transparent_cb.setChecked(bool(self.model.transparent_mask))
        self.widget.autoprocess_cb.setChecked(bool(self.model.img_model.autoprocess))
        self.widget.calibration_lbl.setText(self.model.calibration_model.calibration_name)

        self.update_img_control_widget()
        self.update_mask_mode()
        self.update_roi_in_gui()

        if self.model.current_configuration.auto_integrate_cake and self.widget.img_mode == 'Image':
            self.activate_cake_mode()
        elif not self.model.current_configuration.auto_integrate_cake and self.widget.img_mode == 'Cake':
            self.activate_image_mode()
        elif self.model.current_configuration.auto_integrate_cake and self.widget.img_mode == 'Cake':
            self.set_cake_line_position(self.model.clicked_tth)
            self._update_cake_mouse_click_pos()
        elif not self.model.current_configuration.auto_integrate_cake and self.widget.img_mode == 'Image':
            self._update_image_line_pos()
            self._update_image_mouse_click_pos()

    def change_view_btn_clicked(self):
        if self.view_mode == 'alternative':
            self.change_view_to_normal()
        elif self.view_mode == 'normal':
            self.change_view_to_alternative()

    def change_view_to_normal(self):
        if self.view_mode == 'normal':
            return
        self.vertical_splitter_alternative_state = self.widget.vertical_splitter.saveState()
        self.horizontal_splitter_alternative_state = self.widget.horizontal_splitter.saveState()
        self.widget.vertical_splitter.addWidget(self.widget.integration_pattern_widget)

        self.widget.integration_control_widget.setOrientation(QtCore.Qt.Horizontal)

        if self.vertical_splitter_normal_state:
            self.widget.vertical_splitter.restoreState(self.vertical_splitter_normal_state)
        if self.horizontal_splitter_normal_state:
            self.widget.horizontal_splitter.restoreState(self.horizontal_splitter_normal_state)

        self.widget.img_widget.set_orientation("horizontal")
        self.view_mode = 'normal'

    def change_view_to_alternative(self):
        if self.view_mode == 'alternative':
            return

        self.vertical_splitter_normal_state = self.widget.vertical_splitter.saveState()
        self.horizontal_splitter_normal_state = self.widget.horizontal_splitter.saveState()

        self.widget.vertical_splitter_left.insertWidget(0, self.widget.integration_pattern_widget)

        self.widget.integration_control_widget.setOrientation(QtCore.Qt.Vertical)

        if self.vertical_splitter_alternative_state:
            self.widget.vertical_splitter.restoreState(self.vertical_splitter_alternative_state)
        if self.horizontal_splitter_alternative_state:
            self.widget.horizontal_splitter.restoreState(self.horizontal_splitter_alternative_state)

        self.widget.img_widget.set_orientation("vertical")
        self.view_mode = 'alternative'



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import pathlib
from typing import Optional

from qtpy import QtGui, QtCore, QtWidgets
import pyqtgraph.graphicsItems.GradientEditorItem
import numpy as np

from . import utils
from .NormalizedImageItem import NormalizedImageItem
from ... import style_path


class ColormapPopup(QtWidgets.QFrame):
    """Dialog providing control over the currently used colormap"""

    sigCurrentGradientChanged = QtCore.Signal(dict)
    """Signal emitted when the colormap gradient has changed"""

    sigCurrentNormalizationChanged = QtCore.Signal(str)
    """Signal emitted when the colormap normalization has changed"""

    sigRangeChanged = QtCore.Signal(float, float)
    """Signal emitted when the data range has changed"""

    _RESET_MODES = {  # Button text: (mode, tooltip)
        "Default": ("default", "Use default colormap range autoscale"),
        "Min/max": ("minmax", "Use data min/max to scale colormap range"),
        "Mean±3 Std": ("mean3std", "Use data mean ± 3 × standard deviation to scale colormap range"),
        "Percentile": ("1percentile", "Use data 1st and 99th percentile to scale colormap range"),
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self._data = None
        self.setWindowTitle("Colormap configuration")
        self.setWindowFlags(QtCore.Qt.Popup)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Raised)
        self.setWindowOpacity(0.95)
        self.setLineWidth(2)

        frameLayout = QtWidgets.QVBoxLayout(self)
        frameLayout.setContentsMargins(3, 3, 3, 3)
        frameLayout.setSpacing(3)

        colormapGroupBox = QtWidgets.QGroupBox("Colormap", self)
        frameLayout.addWidget(colormapGroupBox)
        colormapLayout = QtWidgets.QFormLayout(colormapGroupBox)
        colormapLayout.setContentsMargins(0, 6, 0, 0)
        colormapLayout.setLabelAlignment(QtCore.Qt.AlignRight)

        self._gradientComboBox = QtWidgets.QComboBox(self)
        for name, gradient in pyqtgraph.graphicsItems.GradientEditorItem.Gradients.items():
            icon = self._createQIconFromGradient(gradient)
            self._gradientComboBox.addItem(icon, name.capitalize(), gradient)
        self._gradientComboBox.currentIndexChanged.connect(self._gradientComboBoxCurrentIndexChanged)
        colormapLayout.addRow('Colormap:', self._gradientComboBox)

        self._normalizationComboBox = QtWidgets.QComboBox(self)
        for normalization in NormalizedImageItem.supportedNormalizations():
            description = NormalizedImageItem.getNormalizationDescription(normalization).capitalize()
            self._normalizationComboBox.addItem(description, normalization)

        self._normalizationComboBox.setCurrentIndex(0)
        self._normalizationComboBox.currentIndexChanged.connect(self._normalizationComboBoxCurrentIndexChanged)
        colormapLayout.addRow('Normalization:', self._normalizationComboBox)

        rangeGroupBox = QtWidgets.QGroupBox("Range", self)
        frameLayout.addWidget(rangeGroupBox)
        rangeLayout = QtWidgets.QFormLayout(rangeGroupBox)
        rangeLayout.setContentsMargins(0, 6, 0, 0)
        rangeLayout.setLabelAlignment(QtCore.Qt.AlignRight)

        self._minEdit = QtWidgets.QLineEdit(self)
        self._minEdit.setValidator(QtGui.QDoubleValidator(1, float('inf'), -1))
        self._minEdit.editingFinished.connect(self._rangeChanged)
        rangeLayout.addRow('Min:', self._minEdit)

        self._maxEdit = QtWidgets.QLineEdit(self)
        self._maxEdit.setValidator(QtGui.QDoubleValidator(1, float('inf'), -1))
        self._maxEdit.editingFinished.connect(self._rangeChanged)
        rangeLayout.addRow('Max:', self._maxEdit)

        reloadIcon = QtWidgets.QApplication.instance().style().standardIcon(
            QtWidgets.QStyle.SP_BrowserReload
        )
        self._autoscaleButton = QtWidgets.QPushButton(reloadIcon, "Reset", self)
        self._autoscaleButton.setToolTip("Scale colormap range with current mode")
        self._autoscaleButton.clicked.connect(self._autoscaleRequested)
        self._autoscaleButton.setAutoDefault(False)
        self._autoscaleButton.setEnabled(False)
        rangeLayout.addRow("", self._autoscaleButton)

        resetModeGroupBox = QtWidgets.QGroupBox("Reset Mode", self)
        frameLayout.addWidget(resetModeGroupBox)
        resetModesLayout = QtWidgets.QGridLayout(resetModeGroupBox)
        resetModesLayout.setContentsMargins(0, 6, 0, 0)
        resetModesLayout.setSpacing(0)

        self._resetButtonGroup = QtWidgets.QButtonGroup(self)
        for text, (mode, tooltip) in self._RESET_MODES.items():
            radioButton = QtWidgets.QRadioButton(text, self)
            radioButton.setToolTip(tooltip)
            radioButton.setChecked(mode == utils.auto_level.mode)
            self._resetButtonGroup.addButton(radioButton)

        for index, radioButton in enumerate(self._resetButtonGroup.buttons()):
            resetModesLayout.addWidget(radioButton, index // 2, index % 2, QtCore.Qt.AlignLeft)

        self._resetButtonGroup.buttonClicked.connect(self._autoscaleRequested)

        self._filterGapsCheckBox = QtWidgets.QCheckBox(self)
        self._filterGapsCheckBox.setToolTip("Toggle detector gaps value filtering")
        self._filterGapsCheckBox.setChecked(utils.auto_level.filter_dummy)
        self._filterGapsCheckBox.toggled.connect(self._autoscaleRequested)
        nrows = resetModesLayout.rowCount()
        resetModesLayout.addWidget(QtWidgets.QLabel('Filter gaps:'), nrows, 0, QtCore.Qt.AlignRight)
        resetModesLayout.addWidget(self._filterGapsCheckBox, nrows, 1, QtCore.Qt.AlignLeft)

        buttonBox = QtWidgets.QDialogButtonBox(parent=self)
        buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        closeButton = buttonBox.button(QtWidgets.QDialogButtonBox.Close)
        closeButton.clicked.connect(self.close)
        closeButton.setAutoDefault(False)
        frameLayout.addWidget(buttonBox)

    def setData(self, data: Optional[np.ndarray], copy: bool = True):
        """Set data and histogram to use for autoscale"""
        self._data = None if data is None else np.array(data, copy=copy)
        self._autoscaleButton.setEnabled(data is not None)

    def getData(self, copy: bool = True) -> Optional[np.ndarray]:
        """Returns data used for autoscale if set else None"""
        if self._data is None:
            return None
        return np.array(self._data, copy=copy)

    def _gradientComboBoxCurrentIndexChanged(self, index: int):
        if index < 0:
            return
        gradient = self._gradientComboBox.itemData(index, QtCore.Qt.UserRole)
        self.sigCurrentGradientChanged.emit(gradient)

    def _normalizationComboBoxCurrentIndexChanged(self, index: int):
        if index < 0:
            return
        normalization = self._normalizationComboBox.itemData(index, QtCore.Qt.UserRole)
        self.sigCurrentNormalizationChanged.emit(normalization)

    def setCurrentGradient(self, gradient: dict):
        """Set the currently selected gradient

        If the gradient is not available, a 'Custom' item is added for it.
        """
        for name, description in pyqtgraph.graphicsItems.GradientEditorItem.Gradients.items():
            if gradient['mode'] == description['mode'] and gradient['ticks'] == description['ticks']:
                self._gradientComboBox.setCurrentText(name.capitalize())
                return

        icon = self._createQIconFromGradient(gradient)
        # Block signals to avoid emitting with previously selected gradient since index changes
        wasBlocked = self._gradientComboBox.blockSignals(True)
        self._gradientComboBox.insertItem(0, icon, 'Custom', gradient)
        self._gradientComboBox.blockSignals(wasBlocked)
        self._gradientComboBox.setCurrentIndex(0)

    def getCurrentGradient(self) -> dict:
        """Returns the currently selected gradient"""
        return self._gradientComboBox.currentData()

    def setCurrentNormalization(self, normalization: str):
        """Set the currently selected normalization"""
        index = self._normalizationComboBox.findData(normalization)
        if index < 0:
            raise ValueError(f"Unsupported normalization: {normalization}")
        self._normalizationComboBox.setCurrentIndex(index)

    def getCurrentNormalization(self) -> str:
        """Returns the currently selected normalization"""
        return self._normalizationComboBox.currentData()

    def _rangeChanged(self):
        minimum, maximum = self.getRange()
        if maximum < minimum:
            self.setRange(maximum, minimum)
            return
        self.sigRangeChanged.emit(minimum, maximum)

    def setRange(self, minimum: float, maximum: float):
        """Set the data range (min, max) of the colormap"""
        if maximum < minimum:
            minimum, maximum = maximum, minimum
        if np.allclose((minimum, maximum), self.getRange()):
            return
        self._minEdit.setText(self._minEdit.validator().locale().toString(float(minimum)))
        self._maxEdit.setText(self._maxEdit.validator().locale().toString(float(maximum)))
        self._rangeChanged()

    def getRange(self) -> tuple[float, float]:
        """Returns the data range of the colormap (min, max)"""
        minimum, validated = self._minEdit.validator().locale().toDouble(self._minEdit.text())
        if not validated:
            minimum = 1 
        maximum, validated = self._maxEdit.validator().locale().toDouble(self._maxEdit.text())
        if not validated:
            maximum = minimum
        return minimum, maximum

    def _getResetMode(self) -> str:
        button = self._resetButtonGroup.checkedButton()
        if button is not None and button.text() in self._RESET_MODES:
            return self._RESET_MODES[button.text()][0]
        return "default"  # Fallback

    def _autoscaleRequested(self, *args):
        utils.auto_level.mode = self._getResetMode()
        utils.auto_level.filter_dummy = self._filterGapsCheckBox.isChecked()
        colormapRange = utils.auto_level.get_range(self.getData(copy=False))
        if colormapRange is None:
             return
        self.setRange(*colormapRange)

    @staticmethod
    def _createQIconFromGradient(gradient: dict) -> QtGui.QIcon:
        """Generates a QIcon from a pyqtgraph gradient"""
        gradientEditorItem = pyqtgraph.graphicsItems.GradientEditorItem.GradientEditorItem()
        gradientEditorItem.setLength(100)
        gradientEditorItem.restoreState(gradient)
        qgradient = gradientEditorItem.getGradient()

        pixmap = QtGui.QPixmap(100, 100)
        painter = QtGui.QPainter(pixmap)
        brush = QtGui.QBrush(qgradient)
        painter.fillRect(QtCore.QRect(0, 0, 100, 100), brush)
        painter.end()
        return QtGui.QIcon(pixmap)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray data
# Copyright (C) 2019 Clemens Prescher (clemens.prescher@gmail.com)
# Institute for Geology and Mineralogy, University of Cologne
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from qtpy import QtCore, QtWidgets
import pyqtgraph as pq
import pyqtgraph.exporters
import numpy as np
from PIL import Image
import re
import os
from xypattern import Pattern

from .PhotoConfig import gsecars_photo
from ...widgets.MapWidgets import Map2DWidget, SetupMapDialog, OpenBGImageDialog, MapErrorDialog
from ...widgets.UtilityWidgets import save_file_dialog, open_files_dialog
from ...widgets.MainWidget import IntegrationWidget
from ...model.MapModel import MapModel


class MapController(object):
    def __init__(self, working_dir, widget, dioptas_model):
        """
        :param widget: Reference to IntegrationWidget
        :param dioptas_model: Reference to DioptasModel object

        :type widget: IntegrationWidget
        :type dioptas_model: DioptasModel
        :type widget.map_2D_widget: Map2DWidget
        """

        self.working_dir = working_dir
        self.widget = widget  # type: IntegrationWidget
        self.model = dioptas_model  # type: DioptasModel
        self.map_widget = widget.map_2D_widget  # type: Map2DWidget

        self.manual_map_positions_dialog = SetupMapDialog(self.map_widget)
        self.open_bg_image_dialog = OpenBGImageDialog(self.map_widget, gsecars_photo)
        self.map_model = self.model.map_model  # type: MapModel

        self.interactive_mode = True
        self.roi_range = 0.5

        self.bg_image = None
        self.setup_connections()

    def setup_connections(self):
        # Map model Signals
        self.model.map_model.map_changed.connect(self.update_map_image)
        self.model.map_model.map_cleared.connect(self.clear_map)
        self.model.map_model.map_problem.connect(self.map_positions_problem)
        self.model.map_model.roi_problem.connect(self.roi_problem)

        # General Signals
        self.widget.map_btn.clicked.connect(self.map_btn_clicked)
        self.map_widget.widget_raised.connect(self.activate_map)
        self.map_widget.widget_closed.connect(self.deactivate_map)

        self.map_widget.load_map_files_btn.clicked.connect(self.load_map_files_btn_clicked)
        self.map_widget.add_bg_btn.clicked.connect(self.btn_add_bg_image_clicked)
        self.map_widget.bg_opacity_slider.valueChanged.connect(self.modify_map_opacity)
        self.map_widget.snapshot_btn.clicked.connect(self.snapshot_btn_clicked)

        # Pattern widget signals
        self.widget.pattern_widget.map_interactive_roi.sigRegionChanged.connect(self.interactive_roi_range_changed)

        # Map Signals
        self.map_widget.mouse_clicked.connect(self.map_mouse_clicked)
        self.map_widget.mouse_moved.connect(self.map_mouse_moved)

        # ROI
        self.map_widget.roi_add_btn.clicked.connect(self.btn_roi_add_clicked)
        self.map_widget.roi_add_phase_btn.clicked.connect(self.btn_roi_add_phase_clicked)
        self.map_widget.roi_del_btn.clicked.connect(self.btn_roi_del_clicked)
        self.map_widget.roi_clear_btn.clicked.connect(self.btn_roi_clear_clicked)
        self.map_widget.roi_toggle_btn.clicked.connect(self.btn_roi_toggle_clicked)
        self.map_widget.roi_select_all_btn.clicked.connect(self.btn_roi_select_all_clicked)
        self.map_widget.roi_math_txt.textEdited.connect(self.roi_math_txt_changed)
        self.map_widget.roi_list.itemSelectionChanged.connect(self.roi_list_selection_changed)

        # map setup
        self.map_widget.setup_map_btn.clicked.connect(self.manual_map_positions_setup_btn_clicked)
        self.manual_map_positions_dialog.read_list_btn.clicked.connect(self.read_list_btn_clicked)
        self.manual_map_positions_dialog.hor_num_txt.textChanged.connect(self.manual_map_num_points_changed)
        self.manual_map_positions_dialog.ver_num_txt.textChanged.connect(self.manual_map_num_points_changed)
        self.manual_map_positions_dialog.move_up_btn.clicked.connect(self.move_files_up_in_list)
        self.manual_map_positions_dialog.move_down_btn.clicked.connect(self.move_files_down_in_list)
        self.manual_map_positions_dialog.add_empty_btn.clicked.connect(self.add_empty_btn_clicked)
        self.manual_map_positions_dialog.delete_btn.clicked.connect(self.delete_btn_clicked)

    def map_btn_clicked(self):
        if not self.map_widget.isVisible():
            self.map_widget.raise_widget()
        else:
            self.map_widget.close()

    def activate_map(self):
        self.widget.pattern_widget.mouse_left_clicked.connect(self.interactive_roi_pos_changed)
        if self.map_model.is_empty():
            self.load_map_files_btn_clicked()
        else:
            self.widget.pattern_widget.show_map_interactive_roi()
            self.interactive_roi_pos_changed(self.widget.pattern_widget.get_pos_line())

    def deactivate_map(self):
        self.widget.pattern_widget.mouse_left_clicked.disconnect(self.interactive_roi_pos_changed)
        self.widget.pattern_widget.hide_map_interactive_roi()

    def load_map_files_btn_clicked(self):
        filenames = open_files_dialog(self.map_widget, "Load Map files.",
                                      self.model.working_directories['image'])
        if len(filenames):
            self.map_model.reset()

            if os.path.basename(filenames[0]).split('.')[-1] in ['xy', 'chi', 'txt', 'dat']:
                for filename in filenames:
                    self.map_model.add_map_point(filename, Pattern.from_file(filename))

            else:  # assuming the files are image files
                if not self.model.calibration_model.is_calibrated:
                    self.widget.show_error_msg("Can not integrate multiple images without calibration.")
                    return

                self.map_model.map_uses_patterns = False
                self.model.blockSignals()

                progress_dialog = self.widget.get_progress_dialog("Integrating multiple files.", "Abort Integration",
                                                                  len(filenames))

                def callback_fn(current_index):
                    if progress_dialog.wasCanceled():
                        return False
                    progress_dialog.setValue(current_index)
                    QtWidgets.QApplication.processEvents()
                    return ~progress_dialog.wasCanceled()

                self.map_model.load_img_map(filenames, callback_fn)

                    # self.map_model.add_map_point(pattern_path,
                    #                              self.model.pattern,
                    #                              (self.model.img_model.motors_info['Horizontal'],
                    #                               self.model.img_model.motors_info['Vertical']),
                    #                              img_filename=filename)
                self.model.blockSignals(False)
                self.model.img_changed.emit()
                self.model.pattern_changed.emit()

            # now do the necessary steps to update the map
            self.interactive_roi_pos_changed(self.widget.pattern_widget.get_pos_line())
            # self.update_map_status_files_lbl()

    def _integrate_and_save_pattern(self, directory, base_filename):
        self.model.current_configuration.integrate_image_1d()
        path = os.path.join(directory, os.path.splitext(base_filename)[0] + '.xy')
        self.model.current_configuration.save_pattern(path)
        return path

    def _get_map_working_directory(self):
        # if there is no working directory selected A file dialog opens up to choose a directory...
        working_directory = str(QtWidgets.QFileDialog.getExistingDirectory(
            self.map_widget, "Please choose the output directory for the integrated Patterns.",
            self.model.working_directories['pattern']))
        return working_directory

    def interactive_roi_pos_changed(self, line_pos):
        start = line_pos - self.roi_range / 2
        end = line_pos + self.roi_range / 2
        self.update_interactive_roi(start, end)

    def interactive_roi_range_changed(self):
        start = self.widget.pattern_widget.map_interactive_roi.lines[0].value()
        end = self.widget.pattern_widget.map_interactive_roi.lines[1].value()
        self.update_interactive_roi(start, end)

    def update_interactive_roi(self, start, end):
        self.roi_range = end - start
        self.map_model.reset_rois()
        self.map_model.add_roi(start, end, "interactive")
        self.map_model.calculate_map_data()

        if self.widget.pattern_widget.map_interactive_roi not in \
                self.widget.pattern_widget.pattern_plot.items:
            self.widget.pattern_widget.show_map_interactive_roi()
        self.widget.pattern_widget.set_map_interactive_roi(start, end)
        self.update_map_image()

    def map_mouse_clicked(self, x, y):
        position = self.map_model.map.position_from_xy(x, y)
        pattern_filename, img_filename = self.map_model.map.filenames_from_position(position)

        if pattern_filename:
            self.model.pattern_model.load_pattern(pattern_filename)
            self.model.current_configuration.auto_integrate_pattern = False

        if img_filename:
            self.model.img_model.load(img_filename)

        if pattern_filename:
            self.model.current_configuration.auto_integrate_pattern = True

    def map_mouse_moved(self, x, y):
        try:
            position = self.map_model.map.position_from_xy(x, y)
            self.map_widget.map_pos_lbl.setText("{} {}".format(*position))
            pattern_filename, img_filename = self.map_model.map.filenames_from_position(position)
            if pattern_filename:
                self.map_widget.map_filename_lbl.setText("{}".format(os.path.basename(pattern_filename)))
            elif img_filename:
                self.map_widget.map_filename_lbl.setText("{}".format(os.path.basename(img_filename)))

        except AttributeError:
            pass

    def update_map_image(self):
        if self.bg_image is not None:
            map_opacity = self.map_widget.bg_opacity_slider.value()
        else:
            map_opacity = 1.0
        self.map_widget.map_image.setOpacity(map_opacity)
        self.map_widget.map_image.setImage(self.map_model.map.new_image, True)
        self.auto_range()

    # Controls for ROI
    def btn_roi_add_clicked(self, params):
        # roi_num - number of existing ROIs.
        # roi_count - number of rois created (includes deleted rois)
        # calculate ROI position
        try:
            theta_center = params['theta_center']
        except (KeyError, TypeError):
            theta_center = self.map_model.theta_center
        tth_start = theta_center - self.map_model.theta_range
        tth_end = theta_center + self.map_model.theta_range
        roi_start = self.map_model.convert_units(tth_start, '2th_deg', self.map_model.units, self.map_model.wavelength)
        roi_end = self.map_model.convert_units(tth_end, '2th_deg', self.map_model.units, self.map_model.wavelength)

        # add ROI to list
        roi_num = self.map_widget.roi_num
        if roi_num > 25:
            MapErrorDialog(too_many_rois)
            return
        roi_name = self.generate_roi_name(roi_start, roi_end, roi_num)
        roi_list_item = QtWidgets.QListWidgetItem(self.map_widget.roi_list)
        roi_list_item.setText(roi_name)
        roi_list_item.setSelected(True)
        # self.map_widget.map_roi[roi_num]['roi_name'] = roi_name

        # add ROI to pattern view
        roi_count = self.map_widget.roi_count
        self.map_widget.map_roi[roi_count] = {}
        ov = pq.LinearRegionItem.Vertical
        self.map_widget.map_roi[roi_count]['Obj'] = pq.LinearRegionItem(values=[roi_start, roi_end], orientation=ov,
                                                                        movable=True,
                                                                        brush=pq.mkBrush(color=(255, 0, 255, 100)))
        self.map_widget.map_roi[roi_count]['List_Obj'] = self.map_widget.roi_list.item(
            self.map_widget.roi_list.count() - 1)

        self.widget.pattern_widget.pattern_plot.addItem(self.map_widget.map_roi[roi_count]['Obj'])
        self.map_widget.map_roi[roi_count]['Obj'].sigRegionChangeFinished.connect(self.make_roi_changed(roi_count))
        self.map_widget.map_roi[roi_count]['Obj'].sigRegionChanged.connect(self.make_roi_changing(roi_count))
        self.map_widget.roi_num = self.map_widget.roi_num + 1
        self.map_widget.roi_count = self.map_widget.roi_count + 1
        if self.map_widget.roi_num == 1:
            self.toggle_map_widgets_enable(True)
            if self.map_model.map_organized:
                self.map_widget.update_map_btn.click()

        self.roi_list_selection_changed()

    def btn_roi_add_phase_clicked(self):
        cur_ind = self.widget.phase_widget.get_selected_phase_row()
        if cur_ind < 0:
            return
        phase_lines = self.model.phase_model.get_lines_d(cur_ind)
        for line in phase_lines[0:5]:
            tc = {'theta_center': self.map_model.convert_units(line[0], 'd_A', self.map_model.units,
                                                               self.map_model.wavelength)}
            self.btn_roi_add_clicked(tc)

    # create a function for each ROI when ROI modification is done
    def make_roi_changed(self, curr_map_roi):
        def roi_changed():
            if self.map_widget.auto_update_map_cb.isChecked():
                return
            tth_start, tth_end = self.map_widget.map_roi[curr_map_roi]['Obj'].getRegion()
            row = self.map_widget.roi_list.row(self.map_widget.map_roi[curr_map_roi]['List_Obj'])
            new_roi_name = self.generate_roi_name(tth_start, tth_end, row)
            self.map_widget.roi_list.takeItem(row)
            self.map_widget.roi_list.insertItem(row, new_roi_name)
            # self.map_widget.map_roi[curr_map_roi]['roi_name'] = new_roi_name
            self.map_widget.map_roi[curr_map_roi]['List_Obj'] = self.map_widget.roi_list.item(row)
            self.map_widget.roi_list.item(row).setSelected(True)
            self.btn_update_map_clicked()

        return roi_changed

    # create a function for each ROI when ROI modification is going on
    def make_roi_changing(self, curr_map_roi):
        def roi_changing():
            if not self.map_widget.auto_update_map_cb.isChecked():
                return
            tth_start, tth_end = self.map_widget.map_roi[curr_map_roi]['Obj'].getRegion()
            row = self.map_widget.roi_list.row(self.map_widget.map_roi[curr_map_roi]['List_Obj'])
            new_roi_name = self.generate_roi_name(tth_start, tth_end, row)
            self.map_widget.roi_list.takeItem(row)
            self.map_widget.roi_list.insertItem(row, new_roi_name)
            # self.map_widget.map_roi[curr_map_roi]['roi_name'] = new_roi_name
            self.map_widget.map_roi[curr_map_roi]['List_Obj'] = self.map_widget.roi_list.item(row)
            self.map_widget.roi_list.item(row).setSelected(True)
            self.btn_update_map_clicked()

        return roi_changing

    def roi_list_selection_changed(self):
        selected_rois = self.map_widget.roi_list.selectedItems()
        for roi_item in self.map_widget.map_roi:
            if self.map_widget.map_roi[roi_item]['List_Obj'] in selected_rois:
                self.map_widget.map_roi[roi_item]['Obj'].setBrush(pq.mkBrush(color=(255, 255, 0, 100)))
            else:
                self.map_widget.map_roi[roi_item]['Obj'].setBrush(pq.mkBrush(color=(255, 0, 255, 100)))
        # pq.QtGui.QGuiApplication.processEvents()
        self.widget.repaint()

    def generate_roi_name(self, roi_start, roi_end, roi_num):
        roi_name = chr(roi_num + 65) + '_' + '{:.3f}'.format(roi_start) + '-' + '{:.3f}'.format(roi_end)
        return roi_name

    def update_roi_letters(self):
        for row in range(self.map_widget.roi_list.count()):
            curr_roi = self.map_widget.roi_list.item(row)
            curr_roi.setText(chr(row + 65) + '_' + curr_roi.text().split('_')[1])
            # for key in self.map_widget.map_roi:
            #     if self.map_widget.map_roi[key]['List_Obj'] == curr_roi:
            #         self.map_widget.map_roi[key]['roi_name'] = 1
            #         break

    def btn_roi_del_clicked(self):
        for each_roi in self.map_widget.roi_list.selectedItems():
            for key in self.map_widget.map_roi:
                if self.map_widget.map_roi[key]['List_Obj'] == each_roi:
                    self.widget.pattern_widget.pattern_plot.removeItem(self.map_widget.map_roi[key]['Obj'])
                    del self.map_widget.map_roi[key]
                    break
            self.map_widget.roi_list.takeItem(self.map_widget.roi_list.row(each_roi))
            self.map_widget.roi_num = self.map_widget.roi_num - 1
        if self.map_widget.roi_num == 0:
            self.toggle_map_widgets_enable(False)
        else:
            self.update_roi_letters()

    def btn_roi_clear_clicked(self):
        self.map_widget.roi_list.clear()
        for key in self.map_widget.map_roi:
            self.widget.pattern_widget.pattern_plot.removeItem(self.map_widget.map_roi[key]['Obj'])
        self.map_widget.map_roi.clear()
        self.map_widget.roi_num = 0
        self.toggle_map_widgets_enable(False)

    def btn_roi_toggle_clicked(self):
        if self.map_widget.roi_toggle_btn.isChecked():
            for key in self.map_widget.map_roi:
                self.map_widget.map_roi[key]['Obj'].show()
        else:
            for key in self.map_widget.map_roi:
                self.map_widget.map_roi[key]['Obj'].hide()

    def btn_roi_select_all_clicked(self):
        self.map_widget.roi_list.selectAll()

    def roi_math_txt_changed(self):
        existing_rois = []
        roi_math_txt = str(self.map_widget.roi_math_txt.text()).upper()
        roi_math_txt_rois = re.findall('[A-Z]', roi_math_txt)
        for row in range(self.map_widget.roi_list.count()):
            existing_rois.append(self.map_widget.roi_list.item(row).text().split('_')[0])
        for roi in roi_math_txt_rois:
            if roi not in existing_rois:
                self.map_widget.roi_math_txt.setText(self.map_widget.old_roi_math_txt)
                return
        self.map_widget.old_roi_math_txt = roi_math_txt
        self.map_widget.roi_math_txt.setText(roi_math_txt)
        self.btn_update_map_clicked()

    def snapshot_btn_clicked(self):
        snapshot_filename = save_file_dialog(self.widget, "Save Map Snapshot.",
                                             os.path.join(self.working_dir['pattern'], 'map.png'),
                                             'PNG (*.png);;JPG (*.jpg);;TIF (*.tif)')
        exporter = pq.exporters.ImageExporter(self.map_widget.map_view_box)
        exporter.export(snapshot_filename)

    # Auto-range for map image
    def auto_range(self):
        hist_x, hist_y = self.map_widget.map_histogram_LUT.hist_x, self.map_widget.map_histogram_LUT.hist_y
        min_level = hist_x[0]
        max_level = hist_x[-1]
        self.map_widget.map_histogram_LUT.setLevels(min_level, max_level)

    def convert_all_units(self, previous_unit, new_unit, wavelength):
        # also, use this for converting the range if the file is in another unit.
        self.map_widget.roi_list.selectAll()
        for item in self.map_widget.roi_list.selectedItems():
            roi_full_name = item.text().split('_')
            roi_name = roi_full_name[1].split('-')
            roi_start = self.model.map_model.convert_units(float(roi_name[0]), previous_unit, new_unit, wavelength)
            roi_end = self.model.map_model.convert_units(float(roi_name[1]), previous_unit, new_unit, wavelength)
            roi_new_name = self.generate_roi_name(roi_start, roi_end, ord(roi_full_name[0]) - 65)
            item.setText(roi_new_name)
            for key in self.map_widget.map_roi:
                if self.map_widget.map_roi[key]['List_Obj'] == item:
                    self.map_widget.map_roi[key]['Obj'].setRegion((roi_start, roi_end))
                    break

    def btn_add_bg_image_clicked(self):
        if not self.map_widget.map_loaded:
            MapErrorDialog(no_map_loaded)
            return

        load_name = self.load_bg_image_file()

        if not load_name:
            MapErrorDialog(no_bg_image_selected)
            return

        load_name_file = str(load_name).rsplit('/', 1)[-1]
        loaded_bg_image = Image.open(str(load_name).replace('\\', '/'))
        bg_image_tags = loaded_bg_image.tag

        if 'flip_hor_prefixes' in gsecars_photo:
            if load_name_file.split('_')[0] in gsecars_photo['flip_hor_prefixes'].split(','):
                self.open_bg_image_dialog.hor_flip = True
            else:
                self.open_bg_image_dialog.hor_flip = False
        else:
            self.open_bg_image_dialog.hor_flip = gsecars_photo['flip_hor']

        if 'flip_ver_prefixes' in gsecars_photo:
            if load_name_file.split('_')[0] in gsecars_photo['flip_ver_prefixes'].split(','):
                self.open_bg_image_dialog.ver_flip = True
            else:
                self.open_bg_image_dialog.ver_flip = False
        else:
            self.open_bg_image_dialog.ver_flip = gsecars_photo['flip_ver']

        self.bg_hor_ver = self.get_bg_hor_ver(bg_image_tags)
        if 'Horizontal' in self.bg_hor_ver and 'Vertical' in self.bg_hor_ver:
            self.open_bg_image_dialog.hor_center = float(self.bg_hor_ver['Horizontal'])
            self.open_bg_image_dialog.ver_center = float(self.bg_hor_ver['Vertical'])

        self.open_bg_image_dialog.bg_file_name_lbl.setText(load_name)
        self.open_bg_image_dialog.exec_()
        if not self.open_bg_image_dialog.approved:
            return

        img_px_size_hor = self.open_bg_image_dialog.hor_pixel_size
        img_px_size_ver = self.open_bg_image_dialog.ver_pixel_size
        img_hor_px = self.open_bg_image_dialog.hor_num_pixels
        img_ver_px = self.open_bg_image_dialog.ver_num_pixels
        hor_flip = self.open_bg_image_dialog.hor_flip
        ver_flip = self.open_bg_image_dialog.ver_flip
        bg_hor = self.open_bg_image_dialog.hor_center
        bg_ver = self.open_bg_image_dialog.ver_center

        img_width_mm = img_hor_px * img_px_size_hor
        img_height_mm = img_ver_px * img_px_size_ver

        bg_w_px = img_width_mm / self.map_model.hor_um_per_px
        bg_h_px = img_height_mm / self.map_model.ver_um_per_px

        bg_hor_shift = -(-(bg_hor - img_width_mm / 2.0) + self.map_model.min_hor) / self.map_model.hor_um_per_px + \
                       self.map_model.pix_per_hor / 2
        bg_ver_shift = -(-(bg_ver - img_height_mm / 2.0) + self.map_model.min_ver) / self.map_model.ver_um_per_px + \
                       self.map_model.pix_per_ver / 2

        if hor_flip:
            loaded_bg_image = np.fliplr(loaded_bg_image)
        if ver_flip:
            loaded_bg_image = np.flipud(loaded_bg_image)

        self.bg_image = np.rot90(loaded_bg_image, 3)

        self.map_widget.map_bg_image.setImage(self.bg_image)
        bg_rect = QtCore.QRectF(bg_hor_shift, bg_ver_shift, bg_w_px, bg_h_px)
        self.map_widget.map_bg_image.setRect(bg_rect)
        self.modify_map_opacity()

    def load_bg_image_file(self):
        load_name, _ = QtWidgets.QFileDialog.getOpenFileName(QtWidgets.QFileDialog(),
                                                             'Choose file name for loading background image',
                                                             self.working_dir['image'], 'TIFF Files (*.tif)')
        return load_name

    def get_bg_hor_ver(self, tags):
        result = {}
        useful_tags = ['Horizontal:', 'Vertical:']
        for tag in tags:
            for key in useful_tags:
                if key in str(tags[tag]):
                    k, v = str(tags[tag][0]).split(':')
                    result[str(k)] = str(v)
        return result

    def modify_map_opacity(self):
        opacity = self.map_widget.bg_opacity_slider.value() / 100.0
        self.map_widget.map_image.setOpacity(opacity)
        self.map_widget.map_bg_image.setOpacity(1.0 - opacity)

    def clear_map(self):
        self.manual_map_positions_dialog.selected_map_files.clear()

    def map_positions_problem(self):
        MapErrorDialog(map_positions_bad)

    def roi_problem(self):
        MapErrorDialog(wrong_roi_letter)

    def manual_map_positions_setup_btn_clicked(self):
        self.manual_map_positions_dialog.exec_()
        if self.manual_map_positions_dialog.approved:
            map_file_list = []
            for index in range(self.manual_map_positions_dialog.selected_map_files.count()):
                map_file_list.append(str(self.manual_map_positions_dialog.selected_map_files.item(index).text()))
            self.map_model.add_manual_map_positions(self.manual_map_positions_dialog.hor_minimum,
                                                    self.manual_map_positions_dialog.ver_minimum,
                                                    self.manual_map_positions_dialog.hor_step_size,
                                                    self.manual_map_positions_dialog.ver_step_size,
                                                    self.manual_map_positions_dialog.hor_number,
                                                    self.manual_map_positions_dialog.ver_number,
                                                    self.manual_map_positions_dialog.is_hor_first,
                                                    map_file_list)
            self.update_map_status_positions_lbl()
            self.update_map_status_size_and_step_lbl()

    def read_list_btn_clicked(self):
        self.manual_map_positions_dialog.selected_map_files.clear()
        sorted_datalist = self.map_model.sort_map_files_by_natural_name()
        for item in sorted_datalist:
            self.manual_map_positions_dialog.selected_map_files.addItem(QtWidgets.QListWidgetItem(item))
        self.manual_map_positions_dialog.total_files_lbl.setText(
            str(self.manual_map_positions_dialog.selected_map_files.count()) + ' files')
        self.check_num_points()

    def manual_map_num_points_changed(self):
        try:
            self.manual_map_positions_dialog.total_map_points_lbl.setText(str(
                int(self.manual_map_positions_dialog.hor_num_txt.text()) *
                int(self.manual_map_positions_dialog.ver_num_txt.text())) + ' points')
        except ValueError:
            self.manual_map_positions_dialog.total_map_points_lbl.setText('0 points')
        self.check_num_points()

    def check_num_points(self):
        try:
            num_defined = int(self.manual_map_positions_dialog.hor_num_txt.text()) * \
                          int(self.manual_map_positions_dialog.ver_num_txt.text())
            num_in_list = self.manual_map_positions_dialog.selected_map_files.count()
        except ValueError:
            self.manual_map_positions_dialog.ok_btn.setEnabled(False)
            return

        self.manual_map_positions_dialog.ok_btn.setEnabled(num_defined == num_in_list)

    def move_files_up_in_list(self):
        files_list = self.manual_map_positions_dialog.selected_map_files
        selected_files = self.sort_selected_files(files_list)
        for file_name in selected_files:
            row = files_list.row(file_name)
            if row == 0:
                continue
            current_file_name = files_list.takeItem(row)
            files_list.insertItem(row - 1, current_file_name)
            files_list.item(row - 1).setSelected(True)

    def move_files_down_in_list(self):
        files_list = self.manual_map_positions_dialog.selected_map_files
        selected_files = self.sort_selected_files(files_list)
        for file_name in reversed(selected_files):
            row = files_list.row(file_name)
            if row == files_list.count() - 1:
                continue
            current_file_name = files_list.takeItem(row)
            files_list.insertItem(row + 1, current_file_name)
            files_list.item(row + 1).setSelected(True)

    def sort_selected_files(self, files_list):
        selected_files = files_list.selectedItems()
        if not len(selected_files):
            return []
        temp_dict = {}
        for file_name in selected_files:
            temp_dict[files_list.row(file_name)] = file_name

        temp_index = sorted(temp_dict)
        sorted_files = []
        for index in temp_index:
            sorted_files.append(temp_dict[index])
        return sorted_files

    def add_empty_btn_clicked(self):
        files_list = self.manual_map_positions_dialog.selected_map_files
        selected_files = self.sort_selected_files(files_list)
        if selected_files:
            top_row = files_list.row(selected_files[0])
        else:
            top_row = 0
        files_list.insertItem(top_row, "Empty")
        self.manual_map_positions_dialog.total_files_lbl.setText(
            str(self.manual_map_positions_dialog.selected_map_files.count()) + ' files')
        self.check_num_points()

    def delete_btn_clicked(self):
        files_list = self.manual_map_positions_dialog.selected_map_files
        selected_files = self.sort_selected_files(files_list)
        for file_name in selected_files:
            files_list.takeItem(files_list.row(file_name))
        self.manual_map_positions_dialog.total_files_lbl.setText(
            str(self.manual_map_positions_dialog.selected_map_files.count()) + ' files')
        self.check_num_points()


## Map error Messages ####################
##########################################

no_map_loaded = {
    'icon': 'QtWidgets.QMessageBox.Information',
    'detailed_msg': 'Please follow these steps first:\n1. Load a map\n2. Click "Update Map"',
    'short_msg': 'Map not fully loaded',
    'informative_text': 'See additional info...',
    'window_title': 'Error: No Map Shown',
}

no_bg_image_selected = {
    'icon': 'QtWidgets.QMessageBox.Information',
    'detailed_msg': 'Please make sure to select an image when prompted to, and press "Open"',
    'short_msg': 'No background image selected',
    'informative_text': 'See additional info...',
    'window_title': 'Error: No background image selected',
}

map_positions_bad = {
    'icon': 'QtWidgets.QMessageBox.Information',
    'detailed_msg': 'Either not all map files have been loaded or, metadata does not contain map positions.\nPlease'
                    ' set up the map positions manually by clicking "Setup Map"',
    'short_msg': 'Map positions missing',
    'informative_text': 'See additional info...',
    'window_title': 'Warning: Map positions missing',
}

too_many_rois = {
    'icon': 'QtWidgets.QMessageBox.Information',
    'detailed_msg': 'Too many ranges created. try deleting some of the ranges',
    'short_msg': 'Too many ranges',
    'informative_text': 'See additional info...',
    'window_title': 'Warning: Maximum of 26 ranges allowed',
}

wrong_roi_letter = {
    'icon': 'QtWidgets.QMessageBox.Information',
    'detailed_msg': 'Problem with ROI math',
    'short_msg': 'ROI math error',
    'informative_text': 'See additional info...',
    'window_title': 'ROI Math error',
}

cannot_read_positions_from_image_files = {
    'icon': 'QtWidgets.QMessageBox.Information',
    'detailed_msg': 'Please use the Setup Map button in the Map 2D window to setup the map positions',
    'short_msg': 'Could not read positions from image file meta-data',
    'informative_text': 'See additional info...',
    'window_title': 'Positions Missing',
}



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets
from pyqtgraph import GraphicsLayoutWidget

from .plot_widgets import MaskImgWidget

from .CustomWidgets import NumberTextField, LabelAlignRight, SpinBoxAlignRight, HorizontalSpacerItem, \
    CheckableButton, VerticalSpacerItem, HorizontalLine


class MaskWidget(QtWidgets.QWidget):
    """
    Defines the main structure of the mask widget, which is separated into two parts.
    Mask Display Widget - shows the image and pattern
    Mask Control Widget - shows all the controls on the right side of the widget
    """

    def __init__(self, *args, **kwargs):
        super(MaskWidget, self).__init__(*args, **kwargs)
        self.setObjectName('mask_widget')
        self._layout = QtWidgets.QHBoxLayout()
        self.create_display_widget()
        self.create_control_widget()

        self._layout.addWidget(self._display_widget)
        self._layout.addWidget(self._control_widget)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(6)

        self.style_widgets()
        self.setLayout(self._layout)

    def create_display_widget(self):
        self._display_widget = QtWidgets.QWidget(self)
        self._display_layout = QtWidgets.QVBoxLayout()
        self._display_layout.setContentsMargins(0, 0, 0, 0)
        self.img_layout_widget = GraphicsLayoutWidget()
        self.img_widget = MaskImgWidget(self.img_layout_widget)

        self._display_layout.addWidget(self.img_layout_widget)

        self._status_layout = QtWidgets.QHBoxLayout()
        self._status_layout.addSpacerItem(HorizontalSpacerItem())

        self.pos_lbl = LabelAlignRight('')
        self._status_layout.addWidget(self.pos_lbl)
        self._display_layout.addLayout(self._status_layout)

        self._display_widget.setLayout(self._display_layout)

    def create_control_widget(self):
        self._control_widget = QtWidgets.QWidget(self)
        self._control_layout = QtWidgets.QVBoxLayout(self._control_widget)
        self._control_layout.setSpacing(6)

        self._rb_layout = QtWidgets.QHBoxLayout()
        self.mask_rb = QtWidgets.QRadioButton('mask')
        self.unmask_rb = QtWidgets.QRadioButton('unmask')
        self._rb_layout.addWidget(self.mask_rb)
        self._rb_layout.addWidget(self.unmask_rb)
        self._control_layout.addLayout(self._rb_layout)

        self._control_layout.addWidget(HorizontalLine())

        self._geometry_layout = QtWidgets.QGridLayout()
        self.circle_btn = CheckableButton('Circle')
        self.rectangle_btn = CheckableButton('Rectangle')
        self.point_btn = CheckableButton('Point')
        self.point_size_sb = SpinBoxAlignRight()
        self.polygon_btn = CheckableButton('Polygon')
        self.arc_btn = CheckableButton('Arc')
        self._geometry_layout.addWidget(self.circle_btn, 0, 0)
        self._geometry_layout.addWidget(self.rectangle_btn, 0, 1)
        self._geometry_layout.addWidget(self.point_btn, 1, 0)
        self._geometry_layout.addWidget(self.point_size_sb, 1, 1)
        self._geometry_layout.addWidget(self.polygon_btn, 2, 0)
        self._geometry_layout.addWidget(self.arc_btn, 2, 1)
        self._control_layout.addLayout(self._geometry_layout)

        self._control_layout.addWidget(HorizontalLine())

        self._threshold_layout = QtWidgets.QGridLayout()
        self.above_thresh_btn = QtWidgets.QPushButton('Above Thresh')
        self.below_thresh_btn = QtWidgets.QPushButton('Below Thresh')
        self.above_thresh_txt = NumberTextField('')
        self.below_thresh_txt = NumberTextField('')
        self._threshold_layout.addWidget(self.above_thresh_btn, 0, 0)
        self._threshold_layout.addWidget(self.above_thresh_txt, 0, 1)
        self._threshold_layout.addWidget(self.below_thresh_btn, 1, 0)
        self._threshold_layout.addWidget(self.below_thresh_txt, 1, 1)
        self._control_layout.addLayout(self._threshold_layout)

        self._control_layout.addWidget(HorizontalLine())

        self._action_layout = QtWidgets.QGridLayout()
        self.grow_btn = QtWidgets.QPushButton('Grow')
        self.shrink_btn = QtWidgets.QPushButton('Shrink')
        self.invert_mask_btn = QtWidgets.QPushButton('Invert')
        self.clear_mask_btn = QtWidgets.QPushButton('Clear')
        self.undo_btn = QtWidgets.QPushButton('Undo')
        self.redo_btn = QtWidgets.QPushButton('Redo')
        self._action_layout.addWidget(self.grow_btn, 0, 0)
        self._action_layout.addWidget(self.shrink_btn, 0, 1)
        self._action_layout.addWidget(self.invert_mask_btn, 1, 0)
        self._action_layout.addWidget(self.clear_mask_btn, 1, 1)
        self._action_layout.addWidget(self.undo_btn, 2, 0)
        self._action_layout.addWidget(self.redo_btn, 2, 1)
        self._control_layout.addLayout(self._action_layout)

        self._control_layout.addWidget(HorizontalLine())

        self.cosmic_btn = QtWidgets.QPushButton('Cosmic Removal')
        self._control_layout.addWidget(self.cosmic_btn)

        self._control_layout.addWidget(HorizontalLine())

        self._visibility_widget = QtWidgets.QWidget()
        self._visibility_layout = QtWidgets.QHBoxLayout()
        self.fill_rb = QtWidgets.QRadioButton('Fill')
        self.transparent_rb = QtWidgets.QRadioButton('Transparent')
        self._visibility_layout.addWidget(self.fill_rb)
        self._visibility_layout.addWidget(self.transparent_rb)
        self._visibility_widget.setLayout(self._visibility_layout)
        self._control_layout.addWidget(self._visibility_widget)

        self._control_layout.addSpacerItem(VerticalSpacerItem())

        self._file_layout = QtWidgets.QGridLayout()
        self.save_mask_btn = QtWidgets.QPushButton('Save Mask')
        self.load_mask_btn = QtWidgets.QPushButton('Load Mask')
        self.add_mask_btn = QtWidgets.QPushButton('Add Mask')
        self._file_layout.addWidget(self.save_mask_btn, 0, 0, 1, 2)
        self._file_layout.addWidget(self.load_mask_btn, 1, 0)
        self._file_layout.addWidget(self.add_mask_btn, 1, 1)
        self._control_layout.addLayout(self._file_layout)

        self._control_widget.setLayout(self._control_layout)

    def style_widgets(self):
        self.mask_rb.setChecked(True)
        self.fill_rb.setChecked(True)
        self.point_size_sb.setValue(20)

        self._control_widget.setMinimumWidth(200)
        self._control_widget.setMaximumWidth(200)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import os.path

import numpy as np
from dioptas.model.util.signal import Signal

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .Configuration import Configuration


class MapPointInfo:
    filename: str
    frame_index: int

    def __init__(self, filepath, frame_index=0):
        self.filepath = filepath
        self.filename = os.path.basename(filepath)
        self.frame_index = frame_index


class MapModel2:
    map_changed = Signal()

    # point_integrated is emitted with the index of the integrated point
    # it will be a fractional number, when the image file contains multiple frames
    point_integrated = Signal(float)

    def __init__(self, configuration: "Configuration"):
        """
        Creates a new map-model. The configuration specified will serve as
        integrator for the processed files.
        :param configuration: Configuration to be used for integration
        """
        super().__init__()
        self.configuration = configuration
        self.filepaths = None
        self.point_infos = []
        self.pattern_intensities = None
        self.pattern_x = None
        self.window = None
        self.window_intensities = None
        self.dimension = None
        self.possible_dimensions = None
        self.map = None

    def load(self, filepaths: list[str]):
        """Loads a list of files, integrates them and creates a map"""
        if len(filepaths) == 0:
            raise ValueError("No files to load")

        self.filepaths = filepaths

        self.integrate()

        if self.window is None:
            self.window = get_center_window(self.pattern_x)

        self.window_intensities = get_window_intensities(
            self.pattern_x, self.pattern_intensities, self.window
        )

        self.possible_dimensions = find_possible_dimensions(
            len(self.window_intensities)
        )

        if self.dimension is None or self.dimension not in self.possible_dimensions:
            self.dimension = self.possible_dimensions[0]

        self.map = create_map(self.window_intensities, self.dimension)
        self.map_changed.emit()

    def integrate(self):
        """Integrates all files in the filepaths list and stores the results"""
        if not self.configuration.calibration_model.is_calibrated:
            raise ValueError("Detector geometry is not calibrated")

        # initialize data structures
        self.pattern_x = []
        self.pattern_intensities = []
        self.point_infos = []

        # disable trimming trailing zeros for integration, otherwise the
        # integration will result in patterns with different length, which
        # will cause problems when creating the map
        trim_trailing_zeros_backup = self.configuration.trim_trailing_zeros
        self.configuration.trim_trailing_zeros = False

        self.configuration.img_model.img_changed.blocked = True

        try:
            self._integrate()
        except Exception as e:
            self._reset()
            raise e
        finally:
            # reset model to previous state
            self.configuration.trim_trailing_zeros = trim_trailing_zeros_backup
            self.configuration.img_model.img_changed.blocked = False

    def _integrate(self):
        for file_ind, filepath in enumerate(self.filepaths):
            self.configuration.img_model.load(filepath)

            for frame_ind in range(self.configuration.img_model.series_max):
                self.configuration.img_model.load_series_img(frame_ind + 1)
                x, y = self.configuration.integrate_image_1d()

                if file_ind == 0:
                    self.pattern_x = x
                else:
                    if len(x) != len(self.pattern_x):
                        raise ValueError(
                            "The integrated patterns have different length, this is not supported"
                        )

                self.point_infos.append(MapPointInfo(filepath, frame_ind))
                self.pattern_intensities.append(y)

                self.point_integrated.emit(
                    file_ind + (frame_ind + 1) / self.configuration.img_model.series_max
                )
        self.pattern_intensities = np.array(self.pattern_intensities)

    def _reset(self):
        self.filepaths = None
        self.point_infos = []
        self.pattern_intensities = None
        self.pattern_x = None
        self.dimension = None
        self.possible_dimensions = None
        self.map = None
        self.map_changed.emit()

    def set_window(self, window: tuple[float, float]):
        """Sets the window in the pattern for generating the map
        :param window: tuple/list of lower value and upper value of the window
        """
        self.window = window
        if self.pattern_x is None:
            return
        self.window_intensities = get_window_intensities(
            self.pattern_x, self.pattern_intensities, self.window
        )
        self.map = create_map(self.window_intensities, self.dimension)
        self.map_changed.emit()

    def set_dimension(self, dimension: tuple[float, float]):
        """Sets the dimension of the map"""
        if dimension not in self.possible_dimensions:
            return
        self.dimension = dimension
        self.map = create_map(self.window_intensities, self.dimension)
        self.map_changed.emit()

    def get_point_info(self, row_index: float, column_index: float) -> MapPointInfo:
        """Returns the point info for the specified row and column index"""
        if self.dimension is None:
            return None
        ind = self.get_point_index(row_index, column_index)
        return self.point_infos[ind]

    def get_point_index(self, row_index: int, column_index: int) -> int:
        """Returns the point index inside the list of integrated images for the specified row and column index"""
        if self.dimension is None:
            return None
        return int(column_index + self.dimension[1] * row_index)

    def get_point_coordinates(self, index: int) -> tuple[int, int]:
        """Returns the row and column index for the specified point index"""
        if self.dimension is None:
            return None
        return divmod(index, self.dimension[1])

    def get_filenames(self) -> list[str]:
        """Returns a list of filenames for the integrated images, it will add the frame index if it is not 0"""
        filenames = []
        for point_info in self.point_infos:
            if point_info.frame_index == 0:
                filenames.append(point_info.filename)
            else:
                filenames.append(f"{point_info.filename}:{point_info.frame_index}")
        return filenames

    def select_point(self, row_index: int, column_index: int):
        """Selects the point at the specified row and column index, will trigger a load of the image through the
        configuration. Thus the image_changed signal will be sent to all listeners"""
        point_ind = self.get_point_index(row_index, column_index)
        if point_ind is None:
            return
        self.select_point_by_index(point_ind)

    def select_point_by_index(self, index: int):
        """Selects the point at the specified index (considering the list of images), will trigger a load of the
        image through the configuration. Thus the image_changed signal will be sent to all listeners
        """
        if index < 0 or index >= len(self.point_infos):
            return
        point_info = self.point_infos[index]
        self.configuration.img_model.load(
            point_info.filepath,
            point_info.frame_index,
        )


def get_center_window(x, window_range=3) -> list[float, float]:
    """
    Estimates a window of [x_min, x_max] centered in the x value list.
    :param x: a numpy array
    :param window_range: the window will be estimated with +- range * x_step
    :return: windows with [x_min, x_max]
    """
    window_center = x[int(len(x) / 2)]
    x_step = np.mean(np.diff(x))
    return [
        window_center - window_range * x_step,
        window_center + window_range * x_step,
    ]


def ind_in_window(x_array, window: tuple[float, float]) -> np.ndarray:
    """
    Gets the indices of a numpy array which are in the window
    :param x_array: a numpy array
    :param window: tuple/list of lower value and upper value of the window
    :return: list of indices
    """
    return np.where((x_array > window[0]) & (x_array < window[1]))[0]


def get_window_intensities(
    pattern_x, intensities, window: tuple[float, float]
) -> np.ndarray:
    """
    Estimates the intensities inside the specified window
    :param pattern_x: a numpy array of x values from the pattern
    :param intensities: a 2D numpy array holding the intensities of all patterns
    :param window: tuple/list of lower value and upper value of the summing window
    :return: an 1D array containing the sum of  intensities inside the window for each pattern
    """
    indices = ind_in_window(pattern_x, window)
    return np.sum(intensities[:, indices], axis=1)


def find_possible_dimensions(num_points: int) -> list[(int, int)]:
    """
    Finds the possible dimension for a map with a given number of points
    :param num_points: number of points for the map
    :return: list of dimension pairs (x-dimension, y-dimension)
    """
    dimension_pairs = []
    for n in range(1, int(np.floor(np.sqrt(num_points + 1))) + 1):
        if num_points % n == 0:
            dim1 = n
            dim2 = num_points // n
            dimension_pairs.append((dim1, dim2))
            if dim1 != dim2:
                dimension_pairs.append((dim2, dim1))
    dimension_pairs.sort(key=lambda x: ((x[0] + x[1]) / 2 - np.sqrt(num_points)) ** 2)
    return dimension_pairs


def create_map(data: np.ndarray, dimension: tuple[int, int]) -> np.ndarray:
    """
    Creates a new map from the given 1D array and specified dimension. It will
    always create a copy of the data.
    :param data: input data for creating the map
    :param dimension: tuple of integers giving the shape of the map
    :return: numpy array with the reorganized data
    """
    new_data = np.copy(data)
    return np.reshape(new_data, dimension)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from qtpy import QtWidgets, QtCore, QtGui
from pyqtgraph import GraphicsLayoutWidget

from ...plot_widgets.ImgWidget import IntegrationImgWidget, IntegrationCakeWidget
from ...CustomWidgets import FlatButton, CheckableFlatButton, HorizontalSpacerItem, SaveIconButton, DarkCheckableFlatButton
from ..CustomWidgets import MouseCurrentAndClickedWidget, MouseUnitCurrentAndClickedWidget
from .. import CLICKED_COLOR


class IntegrationImgDisplayWidget(QtWidgets.QWidget):
    def __init__(self):
        super(IntegrationImgDisplayWidget, self).__init__()

        self.frame = QtWidgets.QWidget()
        self.frame.setObjectName('img_frame')

        self._frame_layout = QtWidgets.QVBoxLayout()
        self._frame_layout.setContentsMargins(0, 0, 0, 0)
        self._frame_layout.setSpacing(0)

        self.img_pg_layout = GraphicsLayoutWidget()
        self.img_view = IntegrationImgWidget(self.img_pg_layout, orientation='horizontal')
        self.cake_pg_layout = GraphicsLayoutWidget()
        self.cake_view = IntegrationCakeWidget(self.cake_pg_layout, orientation='horizontal')
        self._frame_layout.addWidget(self.img_pg_layout)
        self._frame_layout.addWidget(self.cake_pg_layout)
        self.cake_pg_layout.hide()

        self.position_and_unit_widget = QtWidgets.QWidget()
        self.position_and_unit_widget.setObjectName('img_position_and_unit_widget')
        self._position_and_unit_layout = QtWidgets.QHBoxLayout()
        self._position_and_unit_layout.setContentsMargins(0, 0, 0, 0)

        self.mouse_pos_widget = MouseCurrentAndClickedWidget(CLICKED_COLOR)
        self.mouse_unit_widget = MouseUnitCurrentAndClickedWidget(CLICKED_COLOR)

        self._position_and_unit_layout.addWidget(self.mouse_pos_widget)
        self._position_and_unit_layout.addSpacerItem(HorizontalSpacerItem())
        self._position_and_unit_layout.addWidget(self.mouse_unit_widget)

        self.position_and_unit_widget.setLayout(self._position_and_unit_layout)

        self._frame_layout.addWidget(self.position_and_unit_widget)

        self._control_layout = QtWidgets.QHBoxLayout()
        self._control_layout.setContentsMargins(6, 6, 6, 6)
        self._control_layout.setSpacing(6)

        self.save_image_btn = SaveIconButton()
        self.save_image_btn.setToolTip("Save Image")

        self.roi_btn = CheckableFlatButton('ROI')
        self.mode_btn = FlatButton('Cake')
        self.cake_shift_azimuth_sl = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.mask_btn = CheckableFlatButton('Mask')
        self.phases_btn = CheckableFlatButton('Show Phases')
        self.show_background_subtracted_img_btn = CheckableFlatButton('bkg')
        self.transparent_cb = QtWidgets.QCheckBox('trans')
        self.autoscale_btn = CheckableFlatButton('AutoScale')
        self.undock_btn = FlatButton('Undock')

        self._control_layout.addWidget(self.save_image_btn)
        self._control_layout.addWidget(self.roi_btn)
        self._control_layout.addWidget(self.mode_btn)
        self._control_layout.addWidget(self.cake_shift_azimuth_sl)
        self._control_layout.addWidget(self.mask_btn)
        self._control_layout.addWidget(self.transparent_cb)
        self._control_layout.addWidget(self.show_background_subtracted_img_btn)
        self._control_layout.addWidget(self.phases_btn)
        self._control_layout.addSpacerItem(HorizontalSpacerItem())
        self._control_layout.addWidget(self.autoscale_btn)
        self._control_layout.addWidget(self.undock_btn)

        self._frame_layout.addLayout(self._control_layout)
        self.frame.setLayout(self._frame_layout)

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.frame)

        self.setLayout(self._layout)

        self.style_widgets()
        self.set_tooltips()
        self.cake_shift_azimuth_sl.setVisible(False)
        self.show_background_subtracted_img_btn.setVisible(False)

    def style_widgets(self):
        self.setStyleSheet("""
            #img_frame, #img_position_and_unit_widget {
                background: black;
            }
            """)
        self.autoscale_btn.setChecked(True)
        self.phases_btn.setChecked(False)
        self.phases_btn.setVisible(False)
        self.position_and_unit_widget.hide()

        self.save_image_btn.setIconSize(QtCore.QSize(13, 13))
        self.save_image_btn.setWidth(25)

        # btns = [self.roi_btn, self.mode_btn, self.mask_btn, self.transparent_cb, self.show_background_subtracted_img_btn, 
        #         self.autoscale_btn, self.undock_btn]
        
        # for btn in btns:
        #     btn.setFixedHeight(25)

    def set_tooltips(self):
        self.show_background_subtracted_img_btn.setToolTip("Toggle background subtraction")
        self.mask_btn.setToolTip("Toggle mask")
        self.mode_btn.setToolTip("Toggle between image and cake mode")
        self.undock_btn.setToolTip("Undock image to new window")


# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from qtpy import QtWidgets
from pyqtgraph.opengl import GLSurfacePlotItem
from pyqtgraph.opengl import GLViewWidget, GLGridItem
from pyqtgraph import GraphicsLayoutWidget

from .HistogramLUTItem import HistogramLUTItem


class SurfaceWidget(QtWidgets.QWidget):
    iteration_name = ''

    def __init__(self):
        super(SurfaceWidget, self).__init__()

        self.lut_pg_layout = GraphicsLayoutWidget()
        self.pg_layout = GLViewWidget()
        self.pg_layout.setCameraPosition(distance=3)
        self.surf_view_item = None
        self.pressed_key = None
        self.show_range = np.array([0.0, 1.0])
        self.show_scale = np.array([2., 2., 1.])
        self.g_translate = 0
        self.g_pos = 0
        self.marker = 0
        self.marker_color = [1, 0, 0]
        self.marker_size = 5
        self.data = None

        self.create_graphics()

        self._lut_lo = QtWidgets.QVBoxLayout()
        self._lut_lo.setContentsMargins(0, 0, 0, 0)
        self._lut_lo.addWidget(self.lut_pg_layout)

        self._lut_w = QtWidgets.QWidget()
        self._lut_w.setMaximumHeight(80)
        self._lut_w.setLayout(self._lut_lo)

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self._lut_w)
        self._layout.addWidget(self.pg_layout)

        self.img_histogram_LUT_horizontal = HistogramLUTItem()
        self.img_histogram_LUT_horizontal.gradient.loadPreset('jet')

        self.img_histogram_LUT_horizontal.sigLevelsChanged.connect(self.update_color)
        self.img_histogram_LUT_horizontal.sigLevelChangeFinished.connect(self.update_color)
        self.img_histogram_LUT_horizontal.sigLookupTableChanged.connect(self.update_color)

        self.lut_pg_layout.addItem(self.img_histogram_LUT_horizontal, 0, 1)

        self.setLayout(self._layout)

    def create_graphics(self):
        self.back_grid = GLGridItem()
        self.back_grid.rotate(90, 0, 1, 0)
        self.back_grid.setSize(1, 1, 0)
        self.back_grid.setSpacing(1, 0.1, 1)
        self.back_grid.setDepthValue(10)  # draw grid after surfaces since they may be translucent
        self.pg_layout.addItem(self.back_grid)

        self.base_grid = GLGridItem()
        self.base_grid.setSize(26, 4000, 0)
        self.base_grid.setSpacing(1, 100, 1)
        self.base_grid.setDepthValue(10)  # draw grid after surfaces since they may be translucent
        self.pg_layout.addItem(self.base_grid)

        # self.axis = CustomAxis(self.pg_layout)

        self.surf_view_item = GLSurfacePlotItem(z=np.array([[0]]),
                                                colors=np.array([[0, 0, 0, 0]]),
                                                smooth=False)
        self.surf_view_item.setGLOptions('translucent')
        self.pg_layout.addItem(self.surf_view_item)

    def update_color(self):
        if self.data is not None:
            colors = self.get_colors(self.data).reshape(-1, 4)
            self.surf_view_item.setData(z=self.data, colors=colors)

    def plot_surface(self, data, start, step):
        self.g_pos = int((self.g_translate - start) / step)
        colors = self.get_colors(data).reshape(-1, 4)

        abs_range = self.show_range * (np.nanmax(data) - np.nanmin(data)) + np.nanmin(data)
        self.data = np.copy(data)
        self.data[self.data > abs_range[1]] = abs_range[1]
        self.data[self.data < abs_range[0]] = abs_range[0]

        self.surf_view_item.setData(z=self.data, colors=colors)

        self.img_histogram_LUT_horizontal.imageChanged(img_data=self.data)
        self.img_histogram_LUT_horizontal.setLevels(np.nanmin(self.data), np.nanmax(self.data))

        # self.axis.setSize(*self.show_scale)

        self.update_grids(data)

    def update_grids(self, data):
        self.back_grid.setSize(np.nanmax(data), self.data.shape[1], 0)
        self.base_grid.setSize(self.data.shape[0], self.data.shape[1], 0)

        scale = [self.show_scale[0] / data.shape[0],
                 self.show_scale[1] / data.shape[1],
                 self.show_scale[2] / np.nanmax(data)]

        self.surf_view_item.resetTransform()
        self.surf_view_item.translate(-data.shape[0] / 2., -data.shape[1] / 2., 0)
        self.surf_view_item.scale(*scale, local=False)

        self.back_grid.resetTransform()
        self.back_grid.rotate(90, 0, 1, 0)
        self.back_grid.translate(-data.shape[0] / 2, 0, np.nanmax(data) / 2. + np.nanmin(data))
        self.back_grid.scale(*scale, local=False)

        self.base_grid.resetTransform()
        self.base_grid.translate(0, 0, np.nanmin(data))
        self.base_grid.scale(*scale, local=False)

        # self.axis.setSize(*self.show_scale)
        # self.axis.translate(-data.shape[0] / 2, 0, np.nanmax(data) / 2. + np.nanmin(data))
        # self.axis.diff = [self.show_scale[0] * self.g_pos / data.shape[0], 0, 0]

    def get_colors(self, data):
        lut = self.img_histogram_LUT_horizontal.gradient.getLookupTable(256) / 256.

        level = self.img_histogram_LUT_horizontal.getExpLevels()
        # int_data = ((data - np.nanmin(data)) / (np.nanmax(data) - np.nanmin(data)) * 255).astype(int)
        min = np.nanmin(data)
        if level[0] > 1:
            min = level[0]
        int_data = ((data - min) / (level[1] - min) * 255).astype(int)
        int_data[int_data > 255] = 255
        int_data[int_data < 0] = 0
        int_data = np.nan_to_num(int_data)
        int_data[int_data < 0] = 0
        colors_rgb = lut[int_data]
        colors = np.ones((colors_rgb.shape[0], colors_rgb.shape[1], 4))
        colors[..., :3] = colors_rgb

        colors[:, int(self.marker):int(self.marker) + self.marker_size, :3] = self.marker_color
        colors[self.g_pos, :, :3] = self.marker_color
        return colors



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fabio


class FabioLoader:
    def __init__(self, filename):
        """
        Loads an Hdf5 image produced by ESRF
        :param filename: path to the hdf5 file to be loaded
        """

        self.fabio_image = fabio.open(filename)
        self.series_max = self.fabio_image.nframes

    def get_image(self, ind=0):
        return self.fabio_image.get_frame(ind).data[::-1]



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# imports for type hinting in PyCharm -- DO NOT DELETE
from ....model.DioptasModel import DioptasModel
from ....widgets.plot_widgets import PatternWidget


class PhaseInPatternController(object):
    """
    PhaseInPatternController handles all the interaction between the phases and the pattern view.
    """

    def __init__(self, pattern_widget: PatternWidget, dioptas_model: DioptasModel):
        """
        :param integration_widget: Reference to an IntegrationWidget
        :param dioptas_model: reference to DioptasModel object
        """
        self.model = dioptas_model
        self.pattern_widget = pattern_widget

        self.connect()

    def connect(self):
        self.model.phase_model.phase_added.connect(self.add_phase_plot)
        self.model.phase_model.phase_removed.connect(self.pattern_widget.del_phase)

        self.model.phase_model.phase_changed.connect(self.update_phase_lines)
        self.model.phase_model.phase_changed.connect(self.update_phase_legend)
        self.model.phase_model.phase_changed.connect(self.update_phase_color)
        self.model.phase_model.phase_changed.connect(self.update_phase_visible)

        self.model.phase_model.reflection_added.connect(self.reflection_added)
        self.model.phase_model.reflection_deleted.connect(self.reflection_deleted)

        # pattern signals
        self.pattern_widget.view_box.sigRangeChangedManually.connect(
            self.update_all_phase_lines
        )
        self.pattern_widget.pattern_plot.autoBtn.clicked.connect(
            self.update_all_phase_lines
        )
        self.model.pattern_changed.connect(self.pattern_data_changed)

    def add_phase_plot(self):
        """
        Adds a phase to the Pattern Plot
        """
        axis_range = self.pattern_widget.pattern_plot.viewRange()
        x_range = axis_range[0]
        y_range = axis_range[1]
        positions, intensities, baseline = (
            self.model.phase_model.get_rescaled_reflections(
                -1,
                self.model.pattern,
                x_range,
                y_range,
                self.model.calibration_model.wavelength * 1e10,
                self.model.integration_unit,
            )
        )

        self.pattern_widget.add_phase(
            self.model.phase_model.phases[-1].name,
            positions,
            intensities,
            baseline,
            self.model.phase_model.phase_colors[-1],
        )

    def update_phase_lines(self, ind, axis_range=None):
        """
        Updates the intensities of a specific phase with index ind.
        :param ind: Index of the phase
        :param axis_range: list/tuple of visible x_range and y_range -- ((x_min, x_max), (y_min, y_max))
        """
        if axis_range is None:
            axis_range = self.pattern_widget.view_box.viewRange()

        x_range = axis_range[0]
        y_range = axis_range[1]
        positions, intensities, baseline = (
            self.model.phase_model.get_rescaled_reflections(
                ind,
                self.model.pattern,
                x_range,
                y_range,
                self.model.calibration_model.wavelength * 1e10,
                self.model.integration_unit,
            )
        )

        self.pattern_widget.update_phase_intensities(
            ind, positions, intensities, y_range[0]
        )

    def update_all_phase_lines(self):
        for ind in range(len(self.model.phase_model.phases)):
            self.update_phase_lines(ind)

    def pattern_data_changed(self):
        """
        Function is called after the pattern data has changed.
        """
        self.pattern_widget.update_phase_line_visibilities()

    def update_phase_legend(self, ind):
        name = self.model.phase_model.phases[ind].name
        parameter_str = ""
        pressure = self.model.phase_model.phases[ind].params["pressure"]
        temperature = self.model.phase_model.phases[ind].params["temperature"]
        if pressure != 0:
            parameter_str += "{:0.2f} GPa ".format(pressure)
        if temperature != 0 and temperature != 298 and temperature is not None:
            parameter_str += "{:0.2f} K ".format(temperature)
        self.pattern_widget.rename_phase(ind, parameter_str + name)

    def update_phase_color(self, ind):
        self.pattern_widget.set_phase_color(
            ind, self.model.phase_model.phase_colors[ind]
        )

    def update_phase_visible(self, ind):
        if self.model.phase_model.phase_visible[ind]:
            self.pattern_widget.show_phase(ind)
        else:
            self.pattern_widget.hide_phase(ind)

    def reflection_added(self, ind):
        self.pattern_widget.phases[ind].add_line()
        self.update_phase_lines(ind)

    def reflection_deleted(self, phase_ind, reflection_ind):
        self.pattern_widget.phases[phase_ind].delete_line(reflection_ind)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
import datetime
import threading
import subprocess
import shlex
from functools import partial
from sys import platform as _platform

from qtpy import QtWidgets, QtCore

from ..widgets.MainWidget import MainWidget
from ..model.DioptasModel import DioptasModel
from ..widgets.UtilityWidgets import save_file_dialog, open_file_dialog

from . import CalibrationController
from .integration import IntegrationController
from .MaskController import MaskController
from .ConfigurationController import ConfigurationController
from .MapController import MapController

from dioptas import __version__


class MainController(object):
    """
    Creates the main controller for Dioptas. Creates all the data objects and connects them with the other controllers
    """

    def __init__(self, use_settings=True, settings_directory="default", config_file=None):
        """
        :param use_settings: whether to use previously auto saved state of dioptas
        :param settings_directory: directory where the settings are saved
        :param config_file: a json file path with configuration, currently only used for quick_actions
        """
        self.use_settings = use_settings
        self.widget = MainWidget()

        # create data
        if settings_directory == "default":
            self.settings_directory = os.path.join(os.path.expanduser("~"), ".Dioptas")
        else:
            self.settings_directory = settings_directory

        self.model = DioptasModel()

        self.calibration_controller = CalibrationController(
            self.widget.calibration_widget, self.model
        )
        self.mask_controller = MaskController(self.widget.mask_widget, self.model)
        self.integration_controller = IntegrationController(
            self.widget.integration_widget, self.model
        )
        self.map_controller = MapController(self.widget.map_widget, self.model)

        self.calibration_controller.activate()
        self.integration_controller.image_controller.deactivate()
        self.map_controller.deactivate()
        self.mask_controller.deactivate()

        self.configuration_controller = ConfigurationController(
            configuration_widget=self.widget.configuration_widget,
            dioptas_model=self.model,
            controllers=[
                self.calibration_controller,
                self.mask_controller,
                self.integration_controller,
                self,
            ],
        )

        self.create_signals()
        self.update_title()

        if use_settings:
            QtCore.QTimer.singleShot(0, self.load_default_settings)
            self.setup_backup_timer()

        if config_file is not None:
            self.configuration = json.load(open(config_file, "r"))
            self.create_external_actions()

        self.current_tab_index = 0

    def show_window(self):
        """
        Displays the main window on the screen and makes it active.
        """
        self.widget.show()

        if _platform == "darwin":
            self.widget.setWindowState(
                self.widget.windowState() & ~QtCore.Qt.WindowMinimized
                | QtCore.Qt.WindowActive
            )
            self.widget.activateWindow()
            self.widget.raise_()

    def create_signals(self):
        """
        Creates subscriptions for changing tabs and also newly loaded files which will update the title of the main
                window.
        """
        self.widget.closeEvent = self.close_event
        self.widget.show_configuration_menu_btn.toggled.connect(
            self.widget.configuration_widget.setVisible
        )

        self.widget.calibration_mode_btn.toggled.connect(
            self.widget.calibration_widget.setVisible
        )
        self.widget.mask_mode_btn.toggled.connect(self.widget.mask_widget.setVisible)
        self.widget.integration_mode_btn.toggled.connect(
            self.widget.integration_widget.setVisible
        )
        self.widget.map_mode_btn.toggled.connect(self.widget.map_widget.setVisible)

        self.widget.mode_btn_group.buttonToggled.connect(self.tab_changed)

        self.model.img_changed.connect(self.update_title)
        self.model.pattern_changed.connect(self.update_title)

        self.widget.save_btn.clicked.connect(self.save_btn_clicked)
        self.widget.load_btn.clicked.connect(self.load_btn_clicked)
        self.widget.reset_btn.clicked.connect(self.reset_btn_clicked)

    def tab_changed(self):
        """
        Function which is called when a tab has been selected (calibration, mask, or integration). Performs
        needed initialization tasks.
        :return:
        """
        if self.widget.calibration_mode_btn.isChecked():
            ind = 0
        elif self.widget.mask_mode_btn.isChecked():
            ind = 1
        elif self.widget.integration_mode_btn.isChecked():
            ind = 2
        elif self.widget.map_mode_btn.isChecked():
            ind = 3
        else:
            return

        if ind == self.current_tab_index:
            return

        old_index = self.current_tab_index
        self.current_tab_index = ind

        # changing from mask tab will reintegrate the image
        if old_index == 1:  # mask tab
            if self.model.use_mask and self.model.calibration_model.is_calibrated:
                self.model.current_configuration.integrate_image_1d()
                if self.model.current_configuration.auto_integrate_cake:
                    self.model.current_configuration.integrate_image_2d()

        # update the GUI
        if ind == 2:  # integration tab
            self.integration_controller.image_controller.update_image()

        self.activate_mode(ind)
        self.update_image_display_state(old_index, ind)

    def activate_mode(self, mode_ind):
        controllers = [
            self.calibration_controller,
            self.mask_controller,
            self.integration_controller.image_controller,
            self.map_controller
        ]
        for i, controller in enumerate(controllers):
            if i == mode_ind:
                controller.activate()
            else:
                controller.deactivate()

    def update_image_display_state(self, old_index, new_index):
        img_widgets = [
            self.widget.calibration_widget.img_widget,
            self.widget.mask_widget.img_widget,
            self.widget.integration_widget.img_widget,
            self.widget.map_widget.img_plot_widget
        ]
        old_display_state = img_widgets[old_index].get_display_state()
        img_widgets[new_index].set_display_state(*old_display_state)

    def update_title(self):
        """
        Updates the title bar of the main window. The title bar will always show the current version of Dioptas, the
        image or pattern filenames loaded and the current calibration name.
        """
        img_filename = os.path.basename(self.model.img_model.filename)
        pattern_filename = os.path.basename(self.model.pattern.filename)
        calibration_name = self.model.calibration_model.calibration_name
        year = datetime.datetime.now().year
        dioptas_str = "Dioptas " + __version__ + " - © {} C. Prescher".format(year)

        if img_filename == "" and pattern_filename == "":
            self.widget.setWindowTitle(dioptas_str)
            self.widget.integration_widget.img_frame.setWindowTitle(dioptas_str)
            return

        str = ""
        if img_filename != "":
            str += img_filename
        elif img_filename == "" and pattern_filename != "":
            str += pattern_filename
        if not img_filename == pattern_filename and pattern_filename != "":
            str += ", " + pattern_filename
        if calibration_name != "":
            str += ", calibration: " + calibration_name
        str += " | " + dioptas_str 
        self.widget.setWindowTitle(str)
        self.widget.integration_widget.img_frame.setWindowTitle(str)

    def save_default_settings(self):
        if not os.path.exists(self.settings_directory):
            os.mkdir(self.settings_directory)
        self.model.save(os.path.join(self.settings_directory, "config.dio"))

    def load_default_settings(self):
        config_path = os.path.join(self.settings_directory, "config.dio")
        if os.path.isfile(config_path):
            self.show_window()
            if QtWidgets.QMessageBox.Yes == QtWidgets.QMessageBox.question(
                    self.widget,
                    "Recovering previous state.",
                    "Should Dioptas recover your previous Work?",
                    QtWidgets.QMessageBox.Yes,
                    QtWidgets.QMessageBox.No,
            ):
                self.model.load(os.path.join(self.settings_directory, "config.dio"))
            else:
                self.load_directories()

    def setup_backup_timer(self):
        self.backup_timer = QtCore.QTimer(self.widget)
        self.backup_timer.timeout.connect(self.save_default_settings)
        self.backup_timer.setInterval(600000)  # every 10 minutes
        self.backup_timer.start()

    def save_directories(self):
        """
        Currently used working directories for images, spectra, etc. are saved as csv file in the users directory for
        reuse when Dioptas is started again without loading a configuration
        """
        working_directories_path = os.path.join(
            self.settings_directory, "working_directories.json"
        )
        json.dump(self.model.working_directories, open(working_directories_path, "w"))

    def load_directories(self):
        """
        Loads previously used Dioptas directory paths.
        """
        working_directories_path = os.path.join(
            self.settings_directory, "working_directories.json"
        )
        if os.path.exists(working_directories_path):
            self.model.working_directories = json.load(
                open(working_directories_path, "r")
            )

    def close_event(self, ev):
        """
        Intervention of the Dioptas close event to save settings before closing the Program.
        """
        if self.use_settings:
            self.save_default_settings()
            self.save_directories()
        QtWidgets.QApplication.closeAllWindows()
        ev.accept()

    def save_btn_clicked(self):
        try:
            default_file_name = os.path.join(
                self.model.working_directories["project"], "config.dio"
            )
        except (TypeError, KeyError):
            default_file_name = "."
        filename = save_file_dialog(
            self.widget,
            "Save Current Dioptas Project",
            default_file_name,
            filter="Dioptas Project (*.dio)",
        )

        if filename is not None and filename != "":
            self.model.save(filename)
            self.model.working_directories["project"] = os.path.dirname(filename)

    def load_btn_clicked(self):
        try:
            default_file_name = os.path.join(
                self.model.working_directories["project"], "config.dio"
            )
        except (TypeError, KeyError):
            default_file_name = "."
        filename = open_file_dialog(
            self.widget,
            "Load a Dioptas Project",
            default_file_name,
            filter="Dioptas Project (*.dio)",
        )
        if filename is not None and filename != "":
            self.model.load(filename)
            self.model.working_directories["project"] = os.path.dirname(filename)

    def reset_btn_clicked(self):
        if QtWidgets.QMessageBox.Yes == QtWidgets.QMessageBox.question(
                self.widget,
                "Resetting Dioptas.",
                "Do you really want to reset Dioptas?\nAll unsaved work will be lost!",
                QtWidgets.QMessageBox.Yes,
                QtWidgets.QMessageBox.No,
        ):
            self.model.reset()

    def create_external_actions(self):
        self.widget.create_external_actions(self.configuration["external_actions"])
        for action in self.configuration["external_actions"]:
            self.widget.external_action_btns[action["name"]].clicked.connect(
                partial(
                    self.execute_action,
                    action
                )
            )

    def execute_action(self, action):
        command = format(action["command"])
        arguments = action["arguments"]
        img_path = self.model.img_model.filename
        frame_index = self.model.img_model.series_pos

        combined_arguments = f"{arguments} \"{img_path}\" {frame_index}"
        command_str = " ".join([command, combined_arguments])

        # prepare command_str for Popen
        args = shlex.split(command_str)

        def run_command():
            """Run the command with arguments pulse the image file path."""
            subprocess.Popen(args, shell=True)

        threading.Thread(target=run_command).start()

        return command_str



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import h5py
import re


def first(array):
    """  get first element if the only

    :param array: numpy array
    :type array: :class:`numpy.ndarray`
    :returns: first element of the array
    :type array: :obj:`any`
    """
    try:
        if isinstance(array, np.ndarray) and len(array) == 1:
            return array[0]
    except Exception:
        pass
    return array[...]


class LambdaImage:
    def __init__(self, filename=None, file_list=None):
        """
        Loads an image produced by a Lambda detector.
        :param filename: path to the image file to be loaded
        :return: dictionary with image_data, img_data_lambda and series_max, None if unsuccessful
        """
        detector_identifiers = [["/entry/instrument/detector/description", "Lambda"],
                                ["/entry/instrument/detector/description", b"Lambda"]]
        filenumber_list = [1, 2, 3]
        regex_in = r"(.+_m)\d((_part\d+|).nxs)"
        regex_out = r"\g<1>{}\g<2>"
        data_path = "entry/instrument/detector/data"
        module_positions_path = "/entry/instrument/detector/translation/distance"

        if not filename:
            filename = file_list[0]

        try:
            nx_file = h5py.File(filename, "r")
        except OSError:
            raise IOError("not a loadable hdf5 file")

        for identifier in detector_identifiers:
            try:
                if first(nx_file[identifier[0]]) == identifier[1]:
                    break
            except KeyError:
                pass
        else:
            raise IOError("not a lambda image")

        # the image data is spread over multiple files, so we compile a list of them here
        lambda_files = []
        if file_list:
            for f_name in file_list:
                try:
                    lambda_files.append(h5py.File(f_name, "r"))
                except OSError:
                    pass
        else:
            for moduleIndex in filenumber_list:
                try:
                    lambda_files.append(h5py.File(re.sub(regex_in, regex_out.format(moduleIndex), filename), "r"))
                except OSError:
                    pass

        self.file_list = file_list
        self.full_img_data = [imageFile[data_path] for imageFile in lambda_files]
        self.shapes = np.array([module[0].shape for module in self.full_img_data])
        self._module_pos = np.array([np.ravel(nxim[module_positions_path]).astype(int) for nxim in lambda_files])
        self.img_idx = lambda_files[0]['entry/instrument/detector/sequence_number']

        # remove any empty columns/rows to the left or top of the image data or shift any negative rows/columns into the positive
        np.subtract(self._module_pos, self._module_pos[:, 0].min(), self._module_pos, where=[1, 0, 0])
        np.subtract(self._module_pos, self._module_pos[0][1], self._module_pos, where=[0, 1, 0])
        self.series_max = lambda_files[0][data_path].shape[0]

    def get_image(self, image_nr):
        """
        Gets the data for the given image nr and stitches the tiles together
        :param image_nr: position from which to take the image from the image set
        :return: image_data
        """

        tmp = self.shapes + self._module_pos[:, :2][:, ::-1]
        shape = (np.max(tmp[:, 0]), np.max(tmp[:, 1]))
        image = np.zeros(shape)

        for modulenr, moduleImageData in enumerate(self.full_img_data):
            image[self._module_pos[modulenr, 1]:self._module_pos[modulenr, 1] + self.shapes[modulenr][0],
            self._module_pos[modulenr, 0]:self._module_pos[modulenr, 0] + self.shapes[modulenr][1]] = moduleImageData[
                image_nr]

        return image[::-1]
